# -*- coding: utf-8 -*-
"""Module containing implementing the worker for DH group scanning
"""
# import basic stuff

# import own stuff
import tlsmate.dh_numbers as dh
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.server_profile as server_profile
import tlsmate.tls as tls
import tlsmate.utils as utils

# import other stuff


class ScanDhGroups(plg.Worker):
    name = "dh_groups"
    descr = "scan for supported DH groups"
    prio = 30

    def _update_profile(self, version, g_val, p_val):
        version_prof = self.server_profile.get_version_profile(version)
        group = dh.KnownDhGroups.get_known_group(g_val, p_val)
        if group is None:
            size = len(p_val) * 8
            group = dh.DHNumbers(g_val=g_val, p_val=p_val, size=size)
        version_prof.dh_group = server_profile.SPDhGroup(
            name=group.name, size=group.size, g_value=group.g_val, p_value=group.p_val,
        )

    def run(self):
        """Entry point for the worker.

        We don't cover TLS13 here, as the groups are scanned as part of the
        supported groups scan. TLS13 only supports well known groups.
        """
        versions = [
            tls.Version.SSL30,
            tls.Version.TLS10,
            tls.Version.TLS11,
            tls.Version.TLS12,
        ]
        for version in versions:
            prof_vals = self.server_profile.get_profile_values([version], full_hs=False)
            if version in prof_vals.versions:
                self.client.init_profile(profile_values=prof_vals)
                dh_cs = utils.filter_cipher_suites(
                    prof_vals.cipher_suites,
                    key_algo=[
                        tls.KeyExchangeAlgorithm.DHE_DSS,
                        tls.KeyExchangeAlgorithm.DHE_DSS_EXPORT,
                        tls.KeyExchangeAlgorithm.DHE_RSA,
                        tls.KeyExchangeAlgorithm.DHE_RSA_EXPORT,
                        tls.KeyExchangeAlgorithm.DH_ANON,
                    ],
                )
                if dh_cs:
                    self.client.profile.cipher_suites = dh_cs
                    with self.client.create_connection() as conn:
                        conn.send(msg.ClientHello)
                        conn.wait(msg.ServerHello)
                        conn.wait(msg.Certificate, optional=True)
                        ske = conn.wait(msg.ServerKeyExchange)
                        self._update_profile(version, ske.dh.g_val, ske.dh.p_val)



# -*- coding: utf-8 -*-
"""Module for certificates
"""
# import basic stuff
import logging
from typing import Optional, List, Any
import datetime
import builtins

# import own stuff
import tlsmate.cert_utils as cert_utils
import tlsmate.ext as ext
import tlsmate.kdf as kdf
import tlsmate.tls as tls

# import other stuff
from cryptography import x509
from cryptography.x509.oid import NameOID, ExtensionOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import (
    ec,
    rsa,
    dsa,
    ed25519,
    ed448,
)
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.asymmetric import padding


# list of EV OIDs. Sources:
# - https://chromium.googlesource.com/chromium/src/net/+/refs/heads/main/cert/ev_root_ca_metadata.cc  # noqa
# - https://hg.mozilla.org/mozilla-central/file/tip/security/certverifier/ExtendedValidation.cpp  # noqa
_ev_oids = [
    "1.2.156.112559.1.1.6.1",
    "1.2.392.200091.100.721.1",
    "1.2.616.1.113527.2.5.1.1",
    "1.3.159.1.17.1",
    "1.3.171.1.1.10.5.2",
    "1.3.6.1.4.1.13177.10.1.3.10",
    "1.3.6.1.4.1.14777.6.1.1",
    "1.3.6.1.4.1.14777.6.1.2",
    "1.3.6.1.4.1.17326.10.14.2.1.2",
    "1.3.6.1.4.1.17326.10.14.2.2.2",
    "1.3.6.1.4.1.34697.2.1",
    "1.3.6.1.4.1.34697.2.2",
    "1.3.6.1.4.1.34697.2.3",
    "1.3.6.1.4.1.34697.2.4",
    "1.3.6.1.4.1.40869.1.1.22.3",
    "1.3.6.1.4.1.4146.1.1",
    "1.3.6.1.4.1.4788.2.202.1",
    "1.3.6.1.4.1.6334.1.100.1",
    "1.3.6.1.4.1.6449.1.2.1.5.1",
    "1.3.6.1.4.1.782.1.2.1.8.1",
    "1.3.6.1.4.1.7879.13.24.1",
    "1.3.6.1.4.1.8024.0.2.100.1.2",
    "2.16.156.112554.3",
    "2.16.528.1.1003.1.2.7",
    "2.16.578.1.26.1.3.3",
    "2.16.756.1.89.1.2.1.1",
    "2.16.756.5.14.7.4.8",
    "2.16.792.3.0.4.1.1.4",
    "2.16.840.1.114028.10.1.2",
    "2.16.840.1.114404.1.1.2.4.1",
    "2.16.840.1.114412.2.1",
    "2.16.840.1.114413.1.7.23.3",
    "2.16.840.1.114414.1.7.23.3",
    "2.16.840.1.114414.1.7.24.3",
    "2.23.140.1.1",
]


class Certificate(object):
    """Represents a certificate.

    The arguments der and pem are exclusive.

    Arguments:
        der: the certificate in DER-format (raw bytes)
        pem: the certificate in PEM-format
        x509_cert: a certificate object from the cryptography library
        parse: whether the certificate shall be parsed (i.e., all relevant
            data are extracted from the given der/pem structure), or if it shall just
            be stored. In the latter case the certificate will be parsed if a
            property is accessed.
    """

    def __init__(
        self,
        der: Optional[bytes] = None,
        pem: Optional[bytes] = None,
        x509_cert: Optional[x509.Certificate] = None,
        parse: bool = False,
    ) -> None:

        if (der, pem, x509_cert).count(None) != 2:
            raise ValueError("der, pem and x509_cert are exclusive")

        self._bytes = None
        self._pem = None
        self._parsed = None
        self._subject_str = None
        self._self_signed = None
        self.auth_key_id = None
        self.subject_key_id = None
        self.subject_matches: Optional[bool] = None
        self.fingerprint_sha1 = None
        self.fingerprint_sha256 = None
        self.tls12_signature_algorithms = None
        self.tls13_signature_algorithms = None
        self.crl_status = None
        self.ocsp_status = None
        self.issues: List[str] = []
        self.trusted = tls.ScanState.UNDETERMINED
        self.tls_extensions: List[ext.Extension] = []
        self.issuer_cert = None
        self.ocsp_must_staple = tls.ScanState.FALSE
        self.ocsp_must_staple_multi = tls.ScanState.FALSE
        self.extended_validation = tls.ScanState.NA
        self.from_trust_store = False

        if der is not None:
            self._bytes = der
            if parse:
                self._parsed = x509.load_der_x509_certificate(self._bytes)
                self._parse()

        elif pem is not None:
            if isinstance(pem, str):
                pem = pem.encode()

            self._pem = pem
            self._parsed = x509.load_pem_x509_certificate(pem)
            self._parse()

        else:
            self._parsed = x509_cert
            if parse:
                self._parse()

    def __str__(self):
        return self.subject_str

    def __eq__(self, other):
        return self.bytes == other.bytes

    @property
    def subject_str(self):
        """str: The subject name formatted according to RFC 4514.
        """
        if self._subject_str is None:
            self._subject_str = self._parsed.subject.rfc4514_string()
        return self._subject_str

    @property
    def parsed(self):
        """:obj:`cryptography.x509.Certificate`: the x509 certificate object
        """
        if self._parsed is None:
            self._parsed = x509.load_der_x509_certificate(self._bytes)
            self._parse()
        return self._parsed

    @property
    def bytes(self):
        """bytes: the certificate in raw format, i.e. in DER format.
        """
        if self._bytes is None:
            self._bytes = self.parsed.public_bytes(Encoding.DER)
        return self._bytes

    @property
    def pem(self):
        """bytes: the certificate in pem format (it is a binary string!)"""
        if self._pem is None:
            self._pem = self.parsed.public_bytes(Encoding.PEM)
        return self._pem

    def _determine_signature_algorithms(self, public_key):
        """For a given public key provide the compatible signature algorithms.
        """
        if isinstance(public_key, rsa.RSAPublicKey):
            self.tls12_signature_algorithms = [
                tls.SignatureScheme.RSA_PKCS1_SHA1,
                tls.SignatureScheme.RSA_PKCS1_SHA256,
                tls.SignatureScheme.RSA_PKCS1_SHA384,
                tls.SignatureScheme.RSA_PKCS1_SHA512,
                tls.SignatureScheme.RSA_PKCS1_MD5,
                tls.SignatureScheme.RSA_PKCS1_SHA224,
                # Currently, cryptography does not support RSA-PSS-PSS
                # tls.SignatureScheme.RSA_PSS_PSS_SHA256,
                # tls.SignatureScheme.RSA_PSS_PSS_SHA384,
                # tls.SignatureScheme.RSA_PSS_PSS_SHA512,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
            ]
            self.tls13_signature_algorithms = [
                # tls.SignatureScheme.RSA_PSS_PSS_SHA256,
                # tls.SignatureScheme.RSA_PSS_PSS_SHA384,
                # tls.SignatureScheme.RSA_PSS_PSS_SHA512,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
                tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
            ]

        elif isinstance(public_key, dsa.DSAPublicKey):
            self.tls12_signature_algorithms = [
                tls.SignatureScheme.DSA_MD5,
                tls.SignatureScheme.DSA_SHA1,
                tls.SignatureScheme.DSA_SHA224,
                tls.SignatureScheme.DSA_SHA256,
                tls.SignatureScheme.DSA_SHA384,
                tls.SignatureScheme.DSA_SHA512,
            ]
            self.tls13_signature_algorithms = []

        elif isinstance(public_key, ec.EllipticCurvePublicKey):
            size_to_algo = {
                128: tls.SignatureScheme.ECDSA_SHA1,
                224: tls.SignatureScheme.ECDSA_SECP224R1_SHA224,
                256: tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
                384: tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
                512: tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
            }
            sig_scheme = size_to_algo.get(public_key.curve.key_size)
            if sig_scheme is None:
                raise ValueError(
                    f"unknown key size {public_key.curve.key_size} for ECDSA public key"
                )

            self.tls12_signature_algorithms = [sig_scheme]
            if sig_scheme is tls.SignatureScheme.ECDSA_SHA1:
                self.tls13_signature_algorithms = []
            else:
                self.tls13_signature_algorithms = [sig_scheme]

        elif isinstance(public_key, ed25519.Ed25519PublicKey):
            self.tls12_signature_algorithms = [tls.SignatureScheme.ED25519]
            self.tls13_signature_algorithms = [tls.SignatureScheme.ED25519]

        elif isinstance(public_key, ed448.Ed448PublicKey):
            self.tls12_signature_algorithms = [tls.SignatureScheme.ED448]
            self.tls13_signature_algorithms = [tls.SignatureScheme.ED448]

    def _parse(self):
        """Parse the certificate, so that all attributes are set.
        """
        self.fingerprint_sha1 = self._parsed.fingerprint(hashes.SHA1())
        self.fingerprint_sha256 = self._parsed.fingerprint(hashes.SHA256())
        self.signature_algorithm = cert_utils.map_x509_sig_scheme(
            self._parsed.signature_hash_algorithm, self._parsed.signature_algorithm_oid,
        )
        try:
            key_usage = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.KEY_USAGE
            )
            if key_usage.value.digital_signature:
                self._determine_signature_algorithms(self._parsed.public_key())
        except x509.ExtensionNotFound:
            self._determine_signature_algorithms(self._parsed.public_key())

        try:
            self.auth_key_id = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.AUTHORITY_KEY_IDENTIFIER
            ).value.key_identifier

        except x509.ExtensionNotFound:
            self.auth_key_id = None

        try:
            self.subject_key_id = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.SUBJECT_KEY_IDENTIFIER
            ).value.digest

        except x509.ExtensionNotFound:
            self.subject_key_id = None

        try:
            tls_features = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.TLS_FEATURE
            ).value

            if x509.TLSFeatureType.status_request in tls_features:
                self.ocsp_must_staple = tls.ScanState.TRUE

            if x509.TLSFeatureType.status_request_v2 in tls_features:
                self.ocsp_must_staple_multi = tls.ScanState.TRUE

        except x509.ExtensionNotFound:
            pass

        try:
            basic_constr = self._parsed.extensions.get_extension_for_oid(
                ExtensionOID.BASIC_CONSTRAINTS
            ).value
            if not basic_constr.ca:
                try:
                    cert_policies = self._parsed.extensions.get_extension_for_oid(
                        ExtensionOID.CERTIFICATE_POLICIES
                    ).value

                    if any(
                        pol.policy_identifier.dotted_string in _ev_oids
                        for pol in cert_policies
                    ):
                        self.extended_validation = tls.ScanState.TRUE

                    else:
                        self.extended_validation = tls.ScanState.FALSE

                except x509.ExtensionNotFound:
                    pass

        except x509.ExtensionNotFound:
            pass

    def _common_name(self, name):
        """From a given name, extract the common name

        Note, that there might be multiple common names present, in this case
        simple the first one is returned.
        """
        cns = name.get_attributes_for_oid(NameOID.COMMON_NAME)
        if not cns:
            raise tls.UntrustedCertificate(f'no common name for "{self}"')
        return cns[0].value

    @property
    def self_signed(self):
        """bool: Provide an indication if the certificate is self-signed.
        """
        if self._self_signed is None:
            self._self_signed = cert_utils.equal_names(
                self.parsed.subject, self.parsed.issuer
            )
        return self._self_signed

    def mark_untrusted(self, issue: str) -> None:
        """Mark the certificate as untrusted.

        Arguments:
            issue: the error message containing the reason

        """

        self.trusted = tls.ScanState.FALSE
        issue_long = f"certificate {self}: {issue}"
        logging.debug(issue_long)
        self.issues.append(issue)

    def has_valid_period(self, timestamp: datetime.datetime) -> bool:
        """Determines if the period is valid.

        Arguments:
            timestamp: the timestamp to check against

        Returns:
            An indication if the period is valid
        """

        valid = True
        if timestamp < self.parsed.not_valid_before:
            self.mark_untrusted("validity period not yet reached")
            valid = False

        if timestamp > self.parsed.not_valid_after:
            self.mark_untrusted("validity period exceeded")
            valid = False

        return valid

    def has_valid_subject(self, domain: str) -> bool:
        """Validate if the certificate matches the given domain

        It takes the subject and the subject alternative name into account, and
        supports wildcards as well.

        Arguments:
            domain: the domain to check against (normally used in the SNI)

        Returns:
            indication, if the domain name matches the certificate's subject/SAN
        """

        subject_matches = False
        domain = cert_utils.string_prep(domain)
        no_subdomain = cert_utils.remove_subdomain(domain)

        subject_cn = self._common_name(self.parsed.subject)
        if cert_utils.subject_matches(subject_cn, domain, no_subdomain):
            subject_matches = True

        else:
            try:
                subj_alt_names = self.parsed.extensions.get_extension_for_oid(
                    ExtensionOID.SUBJECT_ALTERNATIVE_NAME
                )
                for name in subj_alt_names.value.get_values_for_type(x509.DNSName):
                    if cert_utils.subject_matches(name, domain, no_subdomain):
                        subject_matches = True
                        break

            except x509.ExtensionNotFound:
                pass

        if not subject_matches:
            self.mark_untrusted("subject name does not match")

        self.subject_matches = subject_matches
        return subject_matches

    def _verify_rsa_pkcs(self, signature, data, hash_algo):
        """Verify RSA PKCSv15 signatures
        """

        self.parsed.public_key().verify(
            signature, data, padding.PKCS1v15(), hash_algo()
        )

    def _verify_dsa(self, signature, data, hash_algo):
        """Verify DSA signatures
        """

        self.parsed.public_key().verify(signature, data, hash_algo())

    def _verify_ecdsa(self, signature, data, hash_algo):
        """Verify ECDSA signatures
        """

        self.parsed.public_key().verify(signature, data, ec.ECDSA(hash_algo()))

    def _verify_xcurve(self, signature, data, hash_algo):
        """Verify X25519 and X488 signatures
        """

        self.parsed.public_key().verify(signature, bytes(data))

    def _verify_rsae_pss(self, signature, data, hash_algo):
        """Verify RSA-PSS signatures
        """

        self.parsed.public_key().verify(
            signature,
            data,
            padding.PSS(
                mgf=padding.MGF1(hash_algo()), salt_length=hash_algo.digest_size
            ),
            hash_algo(),
        )

    _sig_schemes = {
        tls.SignatureScheme.RSA_PKCS1_MD5: (_verify_rsa_pkcs, hashes.MD5),
        tls.SignatureScheme.RSA_PKCS1_SHA1: (_verify_rsa_pkcs, hashes.SHA1),
        tls.SignatureScheme.RSA_PKCS1_SHA224: (_verify_rsa_pkcs, hashes.SHA224),
        tls.SignatureScheme.RSA_PKCS1_SHA256: (_verify_rsa_pkcs, hashes.SHA256),
        tls.SignatureScheme.RSA_PKCS1_SHA384: (_verify_rsa_pkcs, hashes.SHA384),
        tls.SignatureScheme.RSA_PKCS1_SHA512: (_verify_rsa_pkcs, hashes.SHA512),
        tls.SignatureScheme.DSA_MD5: (_verify_dsa, hashes.MD5),
        tls.SignatureScheme.DSA_SHA1: (_verify_dsa, hashes.SHA1),
        tls.SignatureScheme.DSA_SHA224: (_verify_dsa, hashes.SHA224),
        tls.SignatureScheme.DSA_SHA256: (_verify_dsa, hashes.SHA256),
        tls.SignatureScheme.DSA_SHA384: (_verify_dsa, hashes.SHA384),
        tls.SignatureScheme.DSA_SHA512: (_verify_dsa, hashes.SHA512),
        tls.SignatureScheme.ECDSA_SHA1: (_verify_ecdsa, hashes.SHA1),
        tls.SignatureScheme.ECDSA_SECP224R1_SHA224: (_verify_ecdsa, hashes.SHA224),
        tls.SignatureScheme.ECDSA_SECP256R1_SHA256: (_verify_ecdsa, hashes.SHA256),
        tls.SignatureScheme.ECDSA_SECP384R1_SHA384: (_verify_ecdsa, hashes.SHA384),
        tls.SignatureScheme.ECDSA_SECP521R1_SHA512: (_verify_ecdsa, hashes.SHA512),
        tls.SignatureScheme.RSA_PSS_PSS_SHA256: (_verify_rsae_pss, hashes.SHA256),
        tls.SignatureScheme.RSA_PSS_PSS_SHA384: (_verify_rsae_pss, hashes.SHA384),
        tls.SignatureScheme.RSA_PSS_PSS_SHA512: (_verify_rsae_pss, hashes.SHA512),
        tls.SignatureScheme.RSA_PSS_RSAE_SHA256: (_verify_rsae_pss, hashes.SHA256),
        tls.SignatureScheme.RSA_PSS_RSAE_SHA384: (_verify_rsae_pss, hashes.SHA384),
        tls.SignatureScheme.RSA_PSS_RSAE_SHA512: (_verify_rsae_pss, hashes.SHA512),
        tls.SignatureScheme.ED25519: (_verify_xcurve, None),
        tls.SignatureScheme.ED448: (_verify_xcurve, None),
    }

    def validate_signature(
        self,
        sig_scheme: tls.SignatureScheme,
        data: builtins.bytes,
        signature: builtins.bytes,
    ) -> None:
        """Validate a signature with a public key from a given certificate.

        Arguments:
            sig_scheme: The signature scheme to use
            data: the bytes for which the signature is to be validated
            signature: the signature

        Raises:
            cryptography.exceptions.InvalidSignature: If the signature does not
            validate.
        """

        sig_params = self._sig_schemes.get(sig_scheme)
        if sig_params is None:
            raise ValueError(f"signature scheme {sig_scheme} not supported")
        sig_params[0](self, signature, data, sig_params[1])

    def validate_cert_signature(self, cert: "Certificate") -> None:
        """Validate the signature within a certificate

        Arguments:
            cert: the certificate for which the signature shall be checked.

        Raises:
            cryptography.exceptions.InvalidSignature: If the signature does not
                validate.
        """

        self.validate_signature(
            cert.signature_algorithm,
            cert.parsed.tbs_certificate_bytes,
            cert.parsed.signature,
        )


def verify_signed_params(
    prefix: bytes,
    params: Any,
    cert: Certificate,
    default_scheme: tls.SignatureScheme,
    version: tls.Version,
) -> None:
    """Verify the signed parameters from a ServerKeyExchange message.

    Arguments:
        prefix: the bytes to prepend to the data
        params: the parameter block from the ServerKeyExchange message
        cert: the certificate used to validate the data
        default_scheme: the default signature scheme to use (if not present in
            the message)
        version: the TLS version. For TLS1.1 and below the signature is
            constructed differently (using SHA1 + MD digests)
    Raises:
        cryptography.exceptions.InvalidSignature: If the signature does not
            validate.
    """

    data = prefix + params.signed_params

    if (
        default_scheme is tls.SignatureScheme.RSA_PKCS1_SHA1
        and version is not tls.Version.TLS12
    ):
        # Digest is a combination of MD5 and SHA1
        digest = kdf.Kdf()
        digest.start_msg_digest()
        digest.set_msg_digest_algo(None)
        digest.update_msg_digest(data)
        hashed1 = digest.current_msg_digest(suspend=True)
        key = cert.parsed.public_key()
        hashed2 = key.recover_data_from_signature(
            params.signature, padding.PKCS1v15(), None
        )
        if hashed1 != hashed2:
            raise InvalidSignature

    else:
        sig_scheme = params.sig_scheme or default_scheme
        cert.validate_signature(sig_scheme, data, params.signature)



# -*- coding: utf-8 -*-
"""Module containing the class implementing the record layer
"""
# import basic stuff
import logging
from typing import Any, Optional

# import own stuff
import tlsmate.config as conf
import tlsmate.pdu as pdu
import tlsmate.record_layer_state as record_layer_state
import tlsmate.recorder as rec
import tlsmate.socket as socket
import tlsmate.structs as structs
import tlsmate.tls as tls

# import other stuff


class RecordLayer(object):
    """Class implementing the record layer.
    """

    def __init__(self, config: conf.Configuration, recorder: rec.Recorder) -> None:
        self._send_buffer = bytearray()
        self._receive_buffer = bytearray()
        self._fragment_max_size = 4 * 4096
        self._write_state: Optional[record_layer_state.RecordLayerState] = None
        self._read_state: Optional[record_layer_state.RecordLayerState] = None
        self._socket = socket.Socket(config=config, recorder=recorder)
        self._flush_each_fragment = False
        self._ssl2 = False

    def _send_fragment(self, rl_msg: structs.RecordLayerMsg, **kwargs: Any) -> None:
        """Protects a fragment and adds it to the send queue.

        Arguments:
            rl_msg: The record layer message to be sent.
        """

        if self._write_state is not None:
            rl_msg = self._write_state.protect_msg(rl_msg, **kwargs)

        self._send_buffer.extend(pdu.pack_uint8(rl_msg.content_type.value))
        self._send_buffer.extend(pdu.pack_uint16(rl_msg.version.value))
        self._send_buffer.extend(pdu.pack_uint16(len(rl_msg.fragment)))
        self._send_buffer.extend(rl_msg.fragment)
        if self._flush_each_fragment:
            self.flush()

    def _fragment(self, rl_msg, **kwargs):
        """Fragments a given message according the maximum fragment size.

        Each fragment is then protected (if applicable) and added to the send queue.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The message to be sent.
        """
        if rl_msg is None or len(rl_msg.fragment) <= self._fragment_max_size:
            self._send_fragment(rl_msg, **kwargs)
            return

        message = rl_msg.fragment
        while len(message) > self._fragment_max_size:
            frag = message[: self._fragment_max_size]
            message = message[self._fragment_max_size :]
            self._send_fragment(
                structs.RecordLayerMsg(
                    content_type=rl_msg.content_type,
                    version=rl_msg.version,
                    fragment=frag,
                )
            )

        if len(message):
            self._send_fragment(
                structs.RecordLayerMsg(
                    content_type=rl_msg.content_type,
                    version=rl_msg.version,
                    fragment=message,
                )
            )

    def send_message(self, message: structs.RecordLayerMsg, **kwargs: Any) -> None:
        """Does everything the record layer needs to do for sending a message.

        The message is fragmented and protected (i.e. encrypted and authenticated) if
        applicable. Compression is not supported.

        Minimal support for SSL2 is provided as well (no fragmentation, no protection).

        The message may result in multiple fragments to be sent. The fragments are
        added to the send queue but actually not sent to the network yet. Use the
        flush method to do so.

        Arguments:
            message: The message to send.
        """

        if message and message.content_type is tls.ContentType.SSL2:
            self._ssl2 = True
            self._send_buffer.extend(pdu.pack_uint16(len(message.fragment) | 0x8000))
            self._send_buffer.extend(message.fragment)

        else:
            self._fragment(message, **kwargs)

    def open_socket(self, l4_addr: structs.TransportEndpoint) -> None:
        """Opens the socket

        Arguments:
            l4_addr : the l4_addr
        """

        self._socket.open_socket(l4_addr)

    def close_socket(self) -> None:
        """Closes the socket. Obviously.
        """

        self._socket.close_socket()

    def flush(self) -> None:
        """Send all fragments in the send queue.

        This function is useful if e.g. multiple handshake messages shall be sent
        in one record layer message.
        """

        self._socket.sendall(self._send_buffer)
        self._send_buffer = bytearray()

    def wait_rl_msg(
        self, timeout: int = 5, **kwargs: Any
    ) -> Optional[structs.RecordLayerMsg]:
        """Wait for a record layer message to be received from the network.

        Arguments:
            timeout: The timeout in seconds to wait for the message. This
                parameter is optional and defaults to 5 seconds.

        Returns:
            A complete record layer message. If a timeout occurs, None is returned.

        Raises:
            ServerMalfunction: If anything went wrong, e.g. message could not be
                authenticated, wrong padding, etc.
        """

        # wait for record layer header
        rl_len = 2 if self._ssl2 else 5
        while len(self._receive_buffer) < rl_len:
            data = self._socket.recv_data(timeout=timeout)
            if data is None or not len(data):
                # TODO: timeout
                return None

            self._receive_buffer.extend(data)

        if self._ssl2:
            content_type = tls.ContentType.SSL2
            version = tls.Version.SSL20
            length, offset = pdu.unpack_uint16(self._receive_buffer, 0)
            if (length & 0x8000) == 0:
                length &= 0x3FFF  # don't evaluate is-escape bit
                offset += 1  # skip padding byte
                rl_len = 3

            else:
                length &= 0x7FFF

        else:
            ct_type, offset = pdu.unpack_uint8(  # type: ignore
                self._receive_buffer, 0
            )
            content_type = tls.ContentType.val2enum(ct_type, alert_on_failure=True)
            vers, offset = pdu.unpack_uint16(  # type: ignore
                self._receive_buffer, offset
            )
            version = tls.Version.val2enum(vers, alert_on_failure=True)
            length, offset = pdu.unpack_uint16(self._receive_buffer, offset)

        while len(self._receive_buffer) < (length + rl_len):
            data = self._socket.recv_data(timeout=timeout)
            if data is None or not len(data):
                # TODO: timeout
                return None

            self._receive_buffer.extend(data)

        # here we have received at least a complete record layer fragment
        fragment = bytes(self._receive_buffer[rl_len : (length + rl_len)])
        self._receive_buffer = self._receive_buffer[(length + rl_len) :]

        rl_msg = structs.RecordLayerMsg(
            content_type=content_type, version=version, fragment=fragment
        )

        if self._read_state is None:
            return rl_msg

        else:
            return self._read_state.unprotect_msg(rl_msg, **kwargs)

    def update_state(self, new_state: structs.StateUpdateParams) -> None:
        """Update the record layer state.

        I.e. sent or received fragments are encrypted and authenticated.

        Arguments:
            new_state (:obj:`tlsmate.structs.StateUpdateParams`): A complete
                state (either a read state or a write state), containing the
                keying material for the symmetric ciphers and other relevant elements.
        """

        state = record_layer_state.RecordLayerState(new_state)
        if new_state.is_write_state:
            self._write_state = state
            state_type = "WRITE"

        else:
            self._read_state = state
            state_type = "READ"

        logging.debug(f"switching record layer state: {state_type}")
        logging.debug(f"{state_type} enc key: {pdu.dump(state._keys.enc)}")
        if state._iv:
            logging.debug(f"{state_type} iv: {pdu.dump(state._iv)}")

        if state._keys.mac:
            logging.debug(f"{state_type} hmac key: {pdu.dump(state._keys.mac)}")



# -*- coding: utf-8 -*-
"""Helper module for generating and executing test cases
"""
# import basic stuff
import logging

# import own stuff
from tlsmate import tls
from tlsmate import msg
from tlsmate.tlssuite import TlsSuiteTester
from tlsmate import utils

# import other stuff


class CipherSuiteTester(TlsSuiteTester):
    """A class which allows to test a specific cipher suite

    A simple test case scenario is implemented with the goal to test a
    full handshake with for a given cipher suite. The server must be an openssl-
    compatible server, as it is expected to receive the openssl-command from a
    simple http request (opensssl s_server -www ...).

    The yaml file for the recorder is named after the cipher suite (done in the
    base class). The cipher suite is specified as a class property.
    Optionally, the version, the supported groups and the signature algorithms
    may be specified as well, otherwise, defaults are used.

    Derived classes MUST define the cipher suite and the path, anything else
    is not required. The test case is recorded by calling
    CipherSuiteTester.record_testcase, and it is replayed by pytest automatically
    (via test_entry method of the TlsSuiteTester class).
    """

    # The name of the yaml file. If None, the name is taken from the
    # cipher suite name
    name = None
    path = None
    cipher_suite = None
    version = tls.Version.TLS12
    supported_groups = [
        tls.SupportedGroups.X25519,
        tls.SupportedGroups.SECP256R1,
        tls.SupportedGroups.SECP384R1,
        tls.SupportedGroups.SECP521R1,
        tls.SupportedGroups.FFDHE2048,
        tls.SupportedGroups.FFDHE4096,
    ]
    signature_algorithms = [
        tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
        tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
        tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
        tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
        tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
        tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
        tls.SignatureScheme.RSA_PKCS1_SHA256,
        tls.SignatureScheme.RSA_PKCS1_SHA384,
        tls.SignatureScheme.RSA_PKCS1_SHA512,
        tls.SignatureScheme.ECDSA_SHA1,
        tls.SignatureScheme.RSA_PKCS1_SHA1,
    ]

    def update_client(self, client):
        """A method which can be overwritten to update the client as needed.
        """
        pass

    def run(self, tlsmate, is_replaying=False):
        """The basic scenario to be recorded or replayed.
        """
        client = tlsmate.client
        client.init_profile()

        client.profile.versions = [self.version]
        client.profile.cipher_suites = [self.cipher_suite]
        client.profile.supported_groups = self.supported_groups
        client.profile.key_shares = self.supported_groups
        client.profile.signature_algorithms = self.signature_algorithms
        self.update_client(client)

        end_of_tc_reached = False
        with client.create_connection() as conn:
            conn.handshake()

            # Cool feature by openssl: if the server is started with the -www
            # option, an HTTP get request will return some information,
            # including the command line used to start the server. We will
            # extract this line and add it to the yaml file.
            # The command openssl_command.py can then be used to read out this
            # information.
            conn.send(msg.AppData(b"GET / HTTP/1.1\n"))
            app_data = conn.wait(msg.AppData)
            while not len(app_data.data):
                app_data = conn.wait(msg.AppData)
            for line in app_data.data.decode("utf-8").split("\n"):
                if line.startswith("s_server"):
                    logging.debug("openssl_command: " + line)
                    conn.recorder.trace(openssl_command=line)
            end_of_tc_reached = True
        assert end_of_tc_reached is True
        return conn

    def entry(self, is_replaying=False):
        """The entry point for the test case.

        Arguments:
            is_replaying (bool): an indication if the test case is recorded or replayed.
        """
        if not is_replaying:
            utils.set_logging_level("debug")
        name = getattr(self, "name")
        if name is not None:
            self.recorder_yaml = self.name
        else:
            self.recorder_yaml = self.cipher_suite.name
        super().entry(is_replaying)



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tlsmate.workers.supported_groups import ScanSupportedGroups
from tlsmate.tlssuite import TlsSuiteTester
from tlsmate.tlssuite import TlsLibrary


groups_tls12 = ["SECP256R1", "SECP384R1", "SECP521R1", "X25519", "X448"]

groups_tls13 = [
    "SECP256R1",
    "SECP384R1",
    "SECP521R1",
    "X25519",
    "X448",
    "FFDHE2048",
    "FFDHE3072",
    "FFDHE4096",
    "FFDHE6144",
    "FFDHE8192",
]


class TestCase(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_in_yaml = "profile_basic_openssl3_0_0_ed25519_ed448"
    sp_out_yaml = "profile_supported_groups_openssl3_0_0_ed25519_ed448"
    recorder_yaml = "recorder_supported_groups_openssl3_0_0_ed25519_ed448"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-ed448 --cert2 server-ed25519 "
        "-- -www"
    )
    library = TlsLibrary.openssl3_0_0

    server = "localhost"

    def check_tls12(self, profile):
        assert profile["extension_supported"] == "TRUE"
        assert profile["server_preference"] == "FALSE"
        assert profile["groups_advertised"] == "NA"
        assert len(profile["groups"]) == len(groups_tls12)
        for a, b in zip(groups_tls12, profile["groups"]):
            assert a == b["name"]

    def check_tls13(self, profile):
        assert profile["extension_supported"] == "TRUE"
        assert profile["server_preference"] == "FALSE"
        assert profile["groups_advertised"] == "TRUE"
        assert len(profile["groups"]) == len(groups_tls13)
        for a, b in zip(groups_tls13, profile["groups"]):
            assert a == b["name"]

    def check_profile(self, profile):
        self.check_tls12(profile["versions"][4]["supported_groups"])
        self.check_tls13(profile["versions"][5]["supported_groups"])

    def run(self, tlsmate, is_replaying):
        server_profile = tlsmate.server_profile
        client = tlsmate.client
        client.init_profile()
        ScanSupportedGroups(tlsmate).run()

        self.check_profile(server_profile.make_serializable())


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tests.cipher_suite_tester import CipherSuiteTester
from tlsmate import tls
from tlsmate.tlssuite import TlsLibrary


class TestCase(CipherSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the CipherSuiteTester class.
    """

    name = "SupportedGroups_SECP224R1"
    path = pathlib.Path(__file__)

    cipher_suite = tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa "
        "-- -www -cipher ALL -curves "
        "sect233k1:sect163r2:sect163k1:sect233r1:sect283k1:sect283r1:sect409k1:"
        "sect409r1:sect571k1:sect571r1:secp224r1:secp256k1:brainpoolP256r1:"
        "brainpoolP384r1:brainpoolP512r1"
    )
    library = TlsLibrary.openssl1_1_1

    # Uncomment the line below if you do not want to use the default version and
    # adapt it to your needs.
    # version = tls.Version.TLS12

    supported_groups = [tls.SupportedGroups.SECP224R1]


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
import logging
from tests.cipher_suite_tester import CipherSuiteTester
from tlsmate import tls
from tlsmate import msg
from tlsmate.tlssuite import TlsLibrary


class TestCase(CipherSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the CipherSuiteTester class.
    """

    name = "Legacy_Renegotiation_Client"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa "
        "-- -www -cipher ALL -legacy_renegotiation"
    )
    library = TlsLibrary.openssl1_1_1

    # Uncomment the line below if you do not want to use the default version and
    # adapt it to your needs.
    # version = tls.Version.TLS12

    def run(self, tlsmate, is_replaying=False):
        client = tlsmate.client
        client.init_profile()

        client.profile.versions = [tls.Version.TLS12]
        client.profile.cipher_suites = [
            tls.CipherSuite.TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384
        ]
        client.profile.supported_groups = [
            tls.SupportedGroups.SECP256R1,
            tls.SupportedGroups.SECP384R1,
            tls.SupportedGroups.SECP521R1,
        ]
        client.profile.signature_algorithms = [tls.SignatureScheme.RSA_PKCS1_SHA256]
        end_of_tc_reached = False
        with client.create_connection() as conn:
            conn.handshake()
            conn.handshake()
            conn.send(msg.AppData(b"GET / HTTP/1.1\n"))
            while True:
                app_data = conn.wait(msg.AppData)
                if len(app_data.data):
                    break
            for line in app_data.data.decode("utf-8").split("\n"):
                if line.startswith("s_server"):
                    logging.debug("openssl_command: " + line)
            end_of_tc_reached = True
        assert end_of_tc_reached is True


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module scanning for supported groups
"""
# import basic stuff
import logging
import abc

# import own stuff
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.server_profile as server_profile
import tlsmate.tls as tls
import tlsmate.utils as utils

# import other stuff


class _Scan(metaclass=abc.ABCMeta):
    def __init__(self, vers_prof, version, testsuite):
        self._version = version
        self._client = testsuite.client
        self._version_prof = vers_prof
        if not hasattr(vers_prof, "supported_groups"):
            vers_prof.supported_groups = server_profile.SPSupportedGroups()
        self._profile_groups = vers_prof.supported_groups

    @abc.abstractmethod
    def _get_cipher_suites(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _get_group_from_server(self, offered_groups):
        raise NotImplementedError

    def _determine_supported_groups(self):
        offered_groups = self._offered_groups[:]
        supported_groups = []
        max_items = 20

        while offered_groups:
            sub_set = offered_groups[:max_items]
            offered_groups = offered_groups[max_items:]

            while sub_set:
                server_group = self._get_group_from_server(sub_set)
                if server_group is None:
                    break
                supported_groups.append(server_group)
                if server_group not in sub_set:
                    self._profile_groups.extension_supported = tls.ScanState.FALSE
                    self._profile_groups.groups = supported_groups
                    return
                sub_set.remove(server_group)

        if not supported_groups:
            raise tls.ScanError(
                "ECDHE cipher suites negotiated, but no groups supported"
            )

        self._profile_groups.extension_supported = tls.ScanState.TRUE
        self._profile_groups.groups = supported_groups

    def _determine_server_preference(self):
        status = None
        groups = self._profile_groups.groups
        if len(groups) > 1:
            ref_group = groups[0]
            groups.append(groups.pop(0))
            server_group = self._get_group_from_server(groups)
            if server_group is not None:
                if server_group is ref_group:
                    status = tls.ScanState.TRUE
                else:
                    status = tls.ScanState.FALSE
            groups.insert(0, groups.pop())
        else:
            status = tls.ScanState.NA
        if status is not None:
            self._profile_groups.server_preference = status

    @abc.abstractmethod
    def _determine_advertised_group(self):
        raise NotImplementedError

    def scan(self, testsuite_name):
        self._client.profile.cipher_suites = self._get_cipher_suites()
        if not self._client.profile.cipher_suites:
            logging.info(f'no (EC)DH cipher suites supported ("{testsuite_name}")')
            return

        self._client.profile.versions = [self._version]
        self._client.profile.signature_algorithms = [
            tls.SignatureScheme.ED25519,
            tls.SignatureScheme.ED448,
            tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
            tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
            tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
            tls.SignatureScheme.RSA_PKCS1_SHA256,
            tls.SignatureScheme.RSA_PKCS1_SHA384,
            tls.SignatureScheme.RSA_PKCS1_SHA512,
            tls.SignatureScheme.ECDSA_SHA1,
            tls.SignatureScheme.RSA_PKCS1_SHA1,
        ]
        self._determine_supported_groups()

        if self._profile_groups.extension_supported is tls.ScanState.TRUE:
            self._determine_server_preference()
            self._determine_advertised_group()


class _TLS12_Scan(_Scan):

    _offered_groups = tls.SupportedGroups.all()

    def _get_cipher_suites(self):
        cipher_suites = self._version_prof.ciphers.cipher_suites
        return utils.filter_cipher_suites(
            cipher_suites,
            key_algo=[
                tls.KeyExchangeAlgorithm.ECDHE_ECDSA,
                tls.KeyExchangeAlgorithm.ECDHE_RSA,
            ],
        )

    def _get_group_from_server(self, offered_groups):
        self._client.profile.supported_groups = offered_groups

        with self._client.create_connection() as conn:
            conn.send(msg.ClientHello)
            conn.wait(msg.ServerHello)
            conn.wait(msg.Certificate, optional=True)
            try:
                msg_ske = conn.wait(msg.ServerKeyExchange)
            except tls.CurveNotSupportedError as exc:
                return exc.curve
            if msg_ske is None or msg_ske.ec is None:
                return None
            return msg_ske.ec.named_curve

        return None

    def _determine_advertised_group(self):
        self._profile_groups.groups_advertised = tls.ScanState.NA


class _TLS13_Scan(_Scan):

    _offered_groups = [
        tls.SupportedGroups.SECP256R1,
        tls.SupportedGroups.SECP384R1,
        tls.SupportedGroups.SECP521R1,
        tls.SupportedGroups.X25519,
        tls.SupportedGroups.X448,
        tls.SupportedGroups.FFDHE2048,
        tls.SupportedGroups.FFDHE3072,
        tls.SupportedGroups.FFDHE4096,
        tls.SupportedGroups.FFDHE6144,
        tls.SupportedGroups.FFDHE8192,
    ]

    def _get_cipher_suites(self):
        return self._version_prof.ciphers.cipher_suites

    def _get_share_from_server(self, offered_groups):
        self._client.profile.supported_groups = offered_groups
        self._client.profile.key_shares = offered_groups

        with self._client.create_connection() as conn:
            conn.send(msg.ClientHello)
            conn.wait(msg.ServerHello)

        if conn.msg.server_hello is None:
            return None

        try:
            key_share_ext = conn.msg.server_hello.get_extension(tls.Extension.KEY_SHARE)
            key_share = key_share_ext.key_shares[0]
            if key_share.group not in offered_groups:
                raise tls.ScanError("selected group was not offered")
        except AttributeError as exc:
            raise tls.ScanError("cannot get selected group from server_hello") from exc

        return key_share

    def _get_group_from_server(self, offered_groups):
        share = self._get_share_from_server(offered_groups)
        if share is not None:
            return share.group
        return None

    def _determine_advertised_group(self):
        status = None
        groups = self._profile_groups.groups
        self._client.profile.supported_groups = groups[:1]
        self._client.profile.key_shares = groups[:1]
        encrypted_extensions = None

        with self._client.create_connection() as conn:
            conn.send(msg.ClientHello)
            conn.wait(msg.ServerHello)
            conn.wait(msg.ChangeCipherSpec, optional=True)
            encrypted_extensions = conn.wait(msg.EncryptedExtensions)

        if encrypted_extensions is not None:
            supported_group_ext = encrypted_extensions.get_extension(
                tls.Extension.SUPPORTED_GROUPS
            )

            if supported_group_ext is None:
                status = tls.ScanState.FALSE
            else:
                advertised_groups = supported_group_ext.supported_groups
                status = tls.ScanState.TRUE

                if self._profile_groups.server_preference is not tls.ScanState.TRUE:
                    if set(advertised_groups) != set(groups):
                        raise tls.ScanError(
                            "server's advertised groups differ from accepted groups"
                        )
        if status is not None:
            self._profile_groups.groups_advertised = status


class ScanSupportedGroups(plg.Worker):
    name = "groups"
    descr = "scan for supported groups"
    prio = 20

    def run(self):

        versions = self.server_profile.get_versions(exclude=[tls.Version.SSL20])
        self.client.alert_on_invalid_cert = False
        for version in versions:
            vers_prof = self.server_profile.get_version_profile(version)
            if vers_prof is not None:
                if version is tls.Version.TLS13:
                    cls = _TLS13_Scan
                else:
                    cls = _TLS12_Scan
                cls(vers_prof, version, self).scan(self.name)



# -*- coding: utf-8 -*-
"""Implements a class to test the dh-groups worker.
"""
import pathlib
from tlsmate.workers.dh_params import ScanDhGroups
from tlsmate.tlssuite import TlsSuiteTester
from tlsmate.tlssuite import TlsLibrary
from tlsmate.tlsmate import TLSMATE_DIR


class TestCase(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_in_yaml = "profile_sig_algos_openssl1_0_2"
    recorder_yaml = "recorder_dh_groups"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa "
        "-- -www -cipher ALL"
    ) + f" -dhparam {str(TLSMATE_DIR)}/utils/dhparam_modp_2048.pem"
    library = TlsLibrary.openssl1_0_2

    server = "localhost"

    def check_profile(self, prof):
        group = prof["versions"][2]["dh_group"]
        assert group["name"] == "RFC3526: 2048-bit MODP Group"
        assert group["size"] == 2048
        assert group["g_value"] == 2

    def run(self, tlsmate, is_replaying):
        server_profile = tlsmate.server_profile
        ScanDhGroups(tlsmate).run()
        self.check_profile(server_profile.make_serializable())


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module defining a class representing the trust store
"""
# import basic stuff
import contextlib
import logging
import os
from typing import List, Any, Optional

# import own stuff
import tlsmate.cert as crt
import tlsmate.cert_utils as cert_utils
import tlsmate.recorder as rec

# import other stuff
import pem
from cryptography.hazmat.primitives.serialization import Encoding


class TrustStore(object):
    """Represents a trust store containing trusted root certificates

    Objects of this class are iterable, yielding the certificates one by one.

    Arguments:
        ca_files (list of file names): A list of files which contain certificates in
            PEM-format.
    """

    def __init__(self, recorder: rec.Recorder) -> None:
        self._recorder = recorder
        self._ca_files: Optional[List[str]] = None
        self._cert_cache: List[crt.Certificate] = []
        self._fingerprint_cache: List[bytes] = []

    def set_ca_files(self, ca_files: List[str]) -> None:
        """Store the CA files containing certs in PEM format

        Arguments:
            ca_files: A list of file names. Each file can contain multiple
                certificates in PEM format.
        """

        if ca_files:
            for ca_file in ca_files:
                logging.debug(f"using {ca_file} as trust store")
            self._ca_files = ca_files

    def __iter__(self):
        """Iterator over all certificates
        """

        for cert in self._cert_cache:
            yield cert

        if self._ca_files:
            with contextlib.redirect_stderr(open(os.devnull, "w")):
                for file_name in self._ca_files:
                    pem_list = pem.parse_file(file_name)
                    for pem_item in pem_list:
                        if not isinstance(pem_item, pem.Certificate):
                            continue
                        yield crt.Certificate(pem=pem_item.as_bytes())

    def add_cert(self, cert: crt.Certificate) -> None:
        """Add a certificate to the trust store if not yet present.

        Arguments:
            The certificate to add
        """

        if cert.fingerprint_sha256 not in self._fingerprint_cache:
            logging.debug(
                f'adding certificate "{cert.parsed.subject.rfc4514_string()}" '
                f"to trust store cache"
            )
            # TODO: resolve type issue
            assert cert.fingerprint_sha256
            self._fingerprint_cache.append(cert.fingerprint_sha256)
            self._cert_cache.append(cert)
            if self._recorder.is_recording():
                cert_pem = cert.parsed.public_bytes(Encoding.DER).hex()
                self._recorder.trace(trust_store=cert_pem)

    def cert_in_trust_store(self, cert: crt.Certificate) -> bool:
        """Checks if a given certificate is present in the trust store.

        Arguments:
            cert: the certificate to check

        Returns:
            True, if the given certificate is present in the trust store
        """

        if self._ca_files is None and not self._cert_cache:
            return False

        for cert2 in self:
            if cert2 == cert:
                self.add_cert(cert2)
                return True

        return False

    def issuer_in_trust_store(self, issuer_name: Any) -> Optional[crt.Certificate]:
        """Returns the certificate for a given issuer name from the trust store.

        Arguments:
            issuer_name: the name of the issuer

        Returns:
            the issuer certificate or None if the certificate is not found.
        """

        for cert in self:
            # TODO: Optimize this, as the issuer_name is string_prepped with
            # always the same result in the loop
            if cert_utils.equal_names(cert.parsed.subject, issuer_name):
                self.add_cert(cert)
                return cert

        return None



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import datetime
import os
from tlsmate.cert_chain import CertChain
from tlsmate.cert import Certificate
from tlsmate.tls import UntrustedCertificate
from tlsmate import tls
from tlsmate import tlsmate as tm
from cryptography.x509 import ocsp, load_pem_x509_certificate, ReasonFlags
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives import hashes, serialization

import pytest
import requests


def init_crl(ca_dir, crl_manager, ca):
    port = 44400
    if "TLSMATE_CA_PORT" in os.environ:
        port = os.environ["TLSMATE_CA_PORT"]

    pem_file = ca_dir / f"crl/{ca}.crl.pem"
    with open(pem_file, "rb") as fd:
        crl = fd.read()
    crl_manager.add_crl(f"http://crl.localhost:{port}/crl/{ca}.crl", pem_crl=crl)


@pytest.fixture
def tlsmate_cert(ca_dir, trust_store_file):
    mate = tm.TlsMate()
    mate.trust_store.set_ca_files([trust_store_file])
    init_crl(ca_dir, mate.crl_manager, "root-rsa")
    init_crl(ca_dir, mate.crl_manager, "root-ecdsa")
    init_crl(ca_dir, mate.crl_manager, "ca-rsa")
    init_crl(ca_dir, mate.crl_manager, "ca-ecdsa")
    mate.config.set("ocsp", False)
    mate.config.set("crl", True)
    return mate


def test_revoked_certificate(
    tlsmate_cert, valid_time, server_revoked_rsa_cert, ca_rsa_cert
):

    chain = CertChain()
    for cert in (server_revoked_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match=f" {tls.CertCrlStatus.REVOKED}"):
        chain.validate(valid_time, "revoked.localhost", True)


def test_certificate_not_yet_valid(tlsmate_cert, server_rsa_cert, ca_rsa_cert):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match=r"validity period.*not yet reached"):
        chain.validate(datetime.datetime(2000, 2, 27), "localhost", True)


def test_certificate_expired(tlsmate_cert, server_rsa_cert, ca_rsa_cert):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match=r"validity period.*exceeded"):
        chain.validate(datetime.datetime(2200, 2, 27), "localhost", True)


def test_dsa_certificate(tlsmate_cert, valid_time, server_dsa_cert, ca_rsa_cert):

    chain = CertChain()
    for cert in (server_dsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", True)
    assert True


def test_ed25519_certificate(
    tlsmate_cert, valid_time, server_ed25519_cert, ca_ecdsa_cert
):

    chain = CertChain()
    for cert in (server_ed25519_cert, ca_ecdsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", True)
    assert True


def test_ed448_certificate(tlsmate_cert, valid_time, server_ed448_cert, ca_ecdsa_cert):

    chain = CertChain()
    for cert in (server_ed448_cert, ca_ecdsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", True)
    assert True


def test_rsa_with_root_certificate(
    tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert, root_rsa_cert
):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert, root_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", True)
    assert True


def test_wrong_sni(tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match=r"subject name does not match"):
        chain.validate(valid_time, "example.com", True)


def test_root_not_last_in_chain(
    tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert, ca_ecdsa_cert, root_rsa_cert
):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert, ca_ecdsa_cert, root_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", True)
    assert True


def test_issuer_mismatch(tlsmate_cert, valid_time, server_rsa_cert, ca_ecdsa_cert):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_ecdsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match="not found in trust store"):
        chain.validate(valid_time, "localhost", True)


def test_signature_invalid_chain(
    tlsmate_cert, valid_time, server_rsa_cert, ca_2nd_rsa_cert
):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_2nd_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match="not found in trust store"):
        chain.validate(valid_time, "localhost", True)


def test_root_in_chain_not_in_truststore(
    tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert, root_rsa_cert
):

    # hard reset of the trust store
    tlsmate_cert.trust_store._ca_files = None

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert, root_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match="self-signed certificate not found in trust store"
    ):
        chain.validate(valid_time, "localhost", True)


def test_root_not_in_chain_not_in_truststore(
    tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert
):

    # hard reset of the trust store
    tlsmate_cert.trust_store._ca_files = None

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match="not found in trust store"):
        chain.validate(valid_time, "localhost", True)


def test_rsa_san(tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "test.localhost", True)
    assert True

    chain.validate(valid_time, "hello.wildcard.localhost", True)
    assert True


def test_certs_not_in_sequence(
    tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert, root_rsa_cert
):

    chain = CertChain()
    for cert in (server_rsa_cert, root_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", True)
    assert "certificates of the chain are not in sequence" in chain.issues[0]


def test_gratuitous_certificate(
    tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert, ca_ecdsa_cert
):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_ecdsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", False)
    assert (
        "gratuitous certificate, not part of trust chain"
        in chain.certificates[1].issues[0]
    )


def test_root_not_in_chain_not_in_truststore_no_exception(
    tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert
):

    # hard reset of the trust store
    tlsmate_cert.trust_store._ca_files = None

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", False)
    assert chain.successful_validation is False


def requests_post_timeout(url, **kwargs):
    raise requests.Timeout


def test_ocsp_status_timeout(
    monkeypatch, tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert
):

    monkeypatch.setattr(requests, "post", requests_post_timeout)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match=r".*connection to OCSP server .* timed out"
    ):
        chain.validate(valid_time, "localhost", True)


def requests_post_exception(url, **kwargs):
    raise ConnectionResetError


def test_ocsp_status_exception(
    monkeypatch, tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert
):

    monkeypatch.setattr(requests, "post", requests_post_exception)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match=r".*connection to OCSP server .* failed"
    ):
        chain.validate(valid_time, "localhost", True)


def test_ocsp_status_cached_chain(
    monkeypatch, tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert
):

    monkeypatch.setattr(requests, "post", requests_post_exception)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match=r".*connection to OCSP server .* failed"
    ):
        chain.validate(valid_time, "localhost", True)

    with pytest.raises(
        UntrustedCertificate, match=r".*cached status for .* is not valid"
    ):
        chain.validate(valid_time, "localhost", True)


class Response(object):
    def __init__(self, content=None):
        self.ok = True
        self.content = content


orig_validate_ocsp = CertChain._valid_ocsp


def validate_ocsp(self, cert, issuer_cert, timestamp):
    if str(cert) == "CN=localhost,O=The TlsMate Company (Server side) RSA,C=DE":
        return orig_validate_ocsp(self, cert, issuer_cert, timestamp)

    else:
        return True


def test_ocsp_status_invalid_signature(
    monkeypatch, tlsmate_cert, server_rsa_cert, ca_rsa_cert, root_rsa_cert, root_rsa_key
):

    now = datetime.datetime.now()

    def requests_posts(url, **kwargs):
        ca_cert = load_pem_x509_certificate(ca_rsa_cert.as_bytes())
        builder = ocsp.OCSPResponseBuilder()
        builder = builder.add_response(
            cert=load_pem_x509_certificate(server_rsa_cert.as_bytes()),
            issuer=ca_cert,
            cert_status=ocsp.OCSPCertStatus.GOOD,
            algorithm=hashes.SHA1(),
            this_update=now,
            next_update=now + datetime.timedelta(days=1),
            revocation_time=None,
            revocation_reason=None,
        ).responder_id(
            ocsp.OCSPResponderEncoding.HASH,
            load_pem_x509_certificate(root_rsa_cert.as_bytes()),
        )
        ocsp_resp = builder.sign(
            serialization.load_pem_private_key(root_rsa_key.as_bytes(), password=None),
            hashes.SHA1(),
        )

        return Response(content=ocsp_resp.public_bytes(Encoding.DER))

    monkeypatch.setattr(requests, "post", requests_posts)
    monkeypatch.setattr(CertChain, "_valid_ocsp", validate_ocsp)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match=r".*OCSP status is SIGNATURE_INVALID"
    ):
        chain.validate(now, "localhost", True)


def test_ocsp_invalid_reponse(
    monkeypatch, tlsmate_cert, server_rsa_cert, ca_rsa_cert,
):

    now = datetime.datetime.now()

    def requests_posts(url, **kwargs):
        ocsp_resp = ocsp.OCSPResponseBuilder.build_unsuccessful(
            ocsp.OCSPResponseStatus.UNAUTHORIZED
        )
        return Response(content=ocsp_resp.public_bytes(Encoding.DER))

    monkeypatch.setattr(requests, "post", requests_posts)
    monkeypatch.setattr(CertChain, "_valid_ocsp", validate_ocsp)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match=r".*OCSP status is INVALID_RESPONSE"
    ):
        chain.validate(now, "localhost", True)


def test_ocsp_status_with_cert(
    monkeypatch, tlsmate_cert, server_rsa_cert, ca_rsa_cert, ca_rsa_key
):

    now = datetime.datetime.now()

    def requests_posts(url, **kwargs):
        ca_cert = load_pem_x509_certificate(ca_rsa_cert.as_bytes())
        builder = ocsp.OCSPResponseBuilder()
        builder = (
            builder.add_response(
                cert=load_pem_x509_certificate(server_rsa_cert.as_bytes()),
                issuer=ca_cert,
                cert_status=ocsp.OCSPCertStatus.GOOD,
                algorithm=hashes.SHA1(),
                this_update=now,
                next_update=now + datetime.timedelta(days=1),
                revocation_time=None,
                revocation_reason=None,
            )
            .certificates([ca_cert])
            .responder_id(ocsp.OCSPResponderEncoding.HASH, ca_cert)
        )
        ocsp_resp = builder.sign(
            serialization.load_pem_private_key(ca_rsa_key.as_bytes(), password=None),
            hashes.SHA1(),
        )

        return Response(content=ocsp_resp.public_bytes(Encoding.DER))

    monkeypatch.setattr(requests, "post", requests_posts)
    monkeypatch.setattr(CertChain, "_valid_ocsp", validate_ocsp)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(now, "localhost", True)
    assert True


def test_ocsp_status_with_invalid_cert(
    monkeypatch,
    tlsmate_cert,
    server_rsa_cert,
    ca_rsa_cert,
    ca_rsa_key,
    server_expired_rsa_cert,
):

    now = datetime.datetime.now()

    def requests_posts(url, **kwargs):
        ca_cert = load_pem_x509_certificate(ca_rsa_cert.as_bytes())
        builder = ocsp.OCSPResponseBuilder()
        builder = (
            builder.add_response(
                cert=load_pem_x509_certificate(server_rsa_cert.as_bytes()),
                issuer=ca_cert,
                cert_status=ocsp.OCSPCertStatus.GOOD,
                algorithm=hashes.SHA1(),
                this_update=now,
                next_update=now + datetime.timedelta(days=1),
                revocation_time=None,
                revocation_reason=None,
            )
            .certificates(
                [load_pem_x509_certificate(server_expired_rsa_cert.as_bytes())]
            )
            .responder_id(ocsp.OCSPResponderEncoding.HASH, ca_cert)
        )
        ocsp_resp = builder.sign(
            serialization.load_pem_private_key(ca_rsa_key.as_bytes(), password=None),
            hashes.SHA1(),
        )
        return Response(content=ocsp_resp.public_bytes(Encoding.DER))

    monkeypatch.setattr(requests, "post", requests_posts)
    monkeypatch.setattr(CertChain, "_valid_ocsp", validate_ocsp)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match=r".*OCSP status is INVALID_ISSUER_CERT"
    ):
        chain.validate(now, "localhost", True)


def test_ocsp_status_invalid_update_time(
    monkeypatch, tlsmate_cert, server_rsa_cert, ca_rsa_cert, ca_rsa_key
):

    now = datetime.datetime.now()

    def requests_posts(url, **kwargs):
        ca_cert = load_pem_x509_certificate(ca_rsa_cert.as_bytes())
        builder = ocsp.OCSPResponseBuilder()
        builder = (
            builder.add_response(
                cert=load_pem_x509_certificate(server_rsa_cert.as_bytes()),
                issuer=ca_cert,
                cert_status=ocsp.OCSPCertStatus.GOOD,
                algorithm=hashes.SHA1(),
                this_update=now + datetime.timedelta(days=1),
                next_update=now + datetime.timedelta(days=1),
                revocation_time=None,
                revocation_reason=None,
            )
            .certificates([ca_cert])
            .responder_id(ocsp.OCSPResponderEncoding.HASH, ca_cert)
        )
        ocsp_resp = builder.sign(
            serialization.load_pem_private_key(ca_rsa_key.as_bytes(), password=None),
            hashes.SHA1(),
        )

        return Response(content=ocsp_resp.public_bytes(Encoding.DER))

    monkeypatch.setattr(requests, "post", requests_posts)
    monkeypatch.setattr(CertChain, "_valid_ocsp", validate_ocsp)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match=r".*OCSP status is INVALID_TIMESTAMP"
    ):
        chain.validate(now, "localhost", True)


def test_ocsp_status_invalid_next_update(
    monkeypatch, tlsmate_cert, server_rsa_cert, ca_rsa_cert, ca_rsa_key
):

    now = datetime.datetime.now()

    def requests_posts(url, **kwargs):
        ca_cert = load_pem_x509_certificate(ca_rsa_cert.as_bytes())
        builder = ocsp.OCSPResponseBuilder()
        builder = (
            builder.add_response(
                cert=load_pem_x509_certificate(server_rsa_cert.as_bytes()),
                issuer=ca_cert,
                cert_status=ocsp.OCSPCertStatus.GOOD,
                algorithm=hashes.SHA1(),
                this_update=now,
                next_update=now - datetime.timedelta(days=1),
                revocation_time=None,
                revocation_reason=None,
            )
            .certificates([ca_cert])
            .responder_id(ocsp.OCSPResponderEncoding.HASH, ca_cert)
        )
        ocsp_resp = builder.sign(
            serialization.load_pem_private_key(ca_rsa_key.as_bytes(), password=None),
            hashes.SHA1(),
        )

        return Response(content=ocsp_resp.public_bytes(Encoding.DER))

    monkeypatch.setattr(requests, "post", requests_posts)
    monkeypatch.setattr(CertChain, "_valid_ocsp", validate_ocsp)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match=r".*OCSP status is INVALID_TIMESTAMP"
    ):
        chain.validate(now, "localhost", True)


def test_ocsp_invalid_response(
    monkeypatch, tlsmate_cert, server_rsa_cert, ca_rsa_cert, ca_rsa_key
):

    now = datetime.datetime.now()

    def requests_posts(url, **kwargs):
        ocsp_resp = ocsp.OCSPResponseBuilder.build_unsuccessful(
            ocsp.OCSPResponseStatus.UNAUTHORIZED
        )
        return Response(content=ocsp_resp.public_bytes(Encoding.DER))

    monkeypatch.setattr(requests, "post", requests_posts)
    monkeypatch.setattr(CertChain, "_valid_ocsp", validate_ocsp)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(
        UntrustedCertificate, match=r".*OCSP status is INVALID_RESPONSE"
    ):
        chain.validate(now, "localhost", True)


def test_ocsp_status_revoked(
    monkeypatch, tlsmate_cert, server_rsa_cert, ca_rsa_cert, ca_rsa_key
):

    now = datetime.datetime.now()

    def requests_posts(url, **kwargs):
        ca_cert = load_pem_x509_certificate(ca_rsa_cert.as_bytes())
        builder = ocsp.OCSPResponseBuilder()
        builder = (
            builder.add_response(
                cert=load_pem_x509_certificate(server_rsa_cert.as_bytes()),
                issuer=ca_cert,
                cert_status=ocsp.OCSPCertStatus.REVOKED,
                algorithm=hashes.SHA1(),
                this_update=now,
                next_update=now + datetime.timedelta(days=1),
                revocation_time=now - datetime.timedelta(days=100),
                revocation_reason=ReasonFlags.key_compromise,
            )
            .certificates([ca_cert])
            .responder_id(ocsp.OCSPResponderEncoding.HASH, ca_cert)
        )
        ocsp_resp = builder.sign(
            serialization.load_pem_private_key(ca_rsa_key.as_bytes(), password=None),
            hashes.SHA1(),
        )

        return Response(content=ocsp_resp.public_bytes(Encoding.DER))

    monkeypatch.setattr(requests, "post", requests_posts)
    monkeypatch.setattr(CertChain, "_valid_ocsp", validate_ocsp)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match=r".*OCSP status is REVOKED"):
        chain.validate(now, "localhost", True)


def test_ocsp_status_unknown(
    monkeypatch, tlsmate_cert, server_rsa_cert, ca_rsa_cert, ca_rsa_key
):

    now = datetime.datetime.now()

    def requests_posts(url, **kwargs):
        ca_cert = load_pem_x509_certificate(ca_rsa_cert.as_bytes())
        builder = ocsp.OCSPResponseBuilder()
        builder = (
            builder.add_response(
                cert=load_pem_x509_certificate(server_rsa_cert.as_bytes()),
                issuer=ca_cert,
                cert_status=ocsp.OCSPCertStatus.UNKNOWN,
                algorithm=hashes.SHA1(),
                this_update=now,
                next_update=now + datetime.timedelta(days=1),
                revocation_time=None,
                revocation_reason=None,
            )
            .certificates([ca_cert])
            .responder_id(ocsp.OCSPResponderEncoding.HASH, ca_cert)
        )
        ocsp_resp = builder.sign(
            serialization.load_pem_private_key(ca_rsa_key.as_bytes(), password=None),
            hashes.SHA1(),
        )

        return Response(content=ocsp_resp.public_bytes(Encoding.DER))

    monkeypatch.setattr(requests, "post", requests_posts)
    monkeypatch.setattr(CertChain, "_valid_ocsp", validate_ocsp)

    tlsmate_cert.config.set("ocsp", True)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match=r".*OCSP status is UNKNOWN"):
        chain.validate(now, "localhost", True)


def test_cert_no_ids(tlsmate_cert, valid_time, server_no_ids_rsa_cert, ca_rsa_cert):

    chain = CertChain()
    for cert in (server_no_ids_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", True)
    assert True


def test_cert_invalid_signature(
    monkeypatch, tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert
):
    def validate_cert_signature(cert):
        raise Exception("bla bla")

    monkeypatch.setattr(Certificate, "validate_cert_signature", validate_cert_signature)

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    with pytest.raises(UntrustedCertificate, match="invalid signature"):
        chain.validate(valid_time, "localhost", True)


def test_cert_untrusted_trust_path(
    tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert, ca_ecdsa_cert
):

    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert, ca_ecdsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.certificates[2].trusted = tls.ScanState.FALSE
    chain.validate(valid_time, "localhost", False)
    assert (
        chain.certificates[2].issues[0]
        == "certificate part of untrusted alternate trust path"
    )


def test_no_crl(tlsmate_cert, valid_time, server_rsa_cert, ca_rsa_cert):

    tlsmate_cert.config.set("crl", False)
    chain = CertChain()
    for cert in (server_rsa_cert, ca_rsa_cert):
        chain.append_pem_cert(cert.as_bytes())

    chain.validate(valid_time, "localhost", False)
    for cert in chain.certificates:
        assert cert.crl_status is tls.CertCrlStatus.UNDETERMINED



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tests.cipher_suite_tester import CipherSuiteTester
from tlsmate import tls
from tlsmate import msg
from tlsmate.tlssuite import TlsLibrary


class TestCase(CipherSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the CipherSuiteTester class.
    """

    path = pathlib.Path(__file__)
    name = "ClientAuth_ECDSA_SHA256_posthandshake"
    cipher_suite = tls.CipherSuite.TLS_AES_128_GCM_SHA256
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa "
        "-- -cipher ALL -client_sigalgs ECDSA+SHA256 -legacy_renegotiation"
    )
    library = TlsLibrary.openssl1_1_1

    # Uncomment the line below if you do not want to use the default version and
    # adapt it to your needs.
    version = tls.Version.TLS13

    def run(self, tlsmate, is_replaying=False):
        """The basic scenario to be recorded or replayed.
        """
        client = tlsmate.client
        client.init_profile()

        client.profile.versions = [self.version]
        client.profile.cipher_suites = [self.cipher_suite]
        client.profile.supported_groups = self.supported_groups
        client.profile.key_shares = self.supported_groups
        client.profile.signature_algorithms = self.signature_algorithms

        end_of_tc_reached = False
        with client.create_connection() as conn:
            conn.handshake()

            self.server_input("c\n", timeout=200)

            conn.wait(msg.CertificateRequest, timeout=15000)
            conn.send(msg.Certificate)
            conn.send(msg.CertificateVerify)
            conn.send(msg.Finished)
            conn.timeout(500)

            end_of_tc_reached = True
        assert end_of_tc_reached is True
        assert conn.handshake_completed is True
        return conn


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module containing the class implementing the record layer
"""
# import basic stuff
import struct
from typing import Any

# import own stuff
import tlsmate.pdu as pdu
import tlsmate.structs as structs
import tlsmate.tls as tls

# import other stuff
from cryptography.hazmat.primitives import hmac, hashes
from cryptography.hazmat.primitives.ciphers import Cipher, modes, aead


class RecordLayerState(object):
    """Class to represent a dynamic record layer state

    Attributes:
        param (:obj:`tlsmate.structs.StateUpdateParams`): the structure to initialize
            the record layer state.
    """

    def __init__(self, param: structs.StateUpdateParams) -> None:

        self._keys = param.keys
        self._mac = param.mac
        self._cipher = param.cipher
        self._compr = param.compr
        self._enc_then_mac = param.enc_then_mac
        self._version = param.version
        self._seq_nbr = 0
        self._cipher_object = None
        self._iv = param.keys.iv
        if self._cipher.c_type == tls.CipherType.STREAM:
            cipher = Cipher(
                self._cipher.algo(self._keys.enc), mode=None  # type: ignore
            )
            if param.is_write_state:
                self._cipher_object = cipher.encryptor()

            else:
                self._cipher_object = cipher.decryptor()

        if self._version is tls.Version.TLS13:
            self._cipher_object = param.cipher.algo

    def _encrypt_cbc(self, fragment, padding_cb=None, **kwargs):
        """Encrypt a fragment using a block cipher in CBC mode.

        Arguments:
            fragment (bytes): The fragment to encrypt.

        Returns:
            bytes: The encrypted fragment.
        """

        # padding
        length = len(fragment) + 1
        missing_bytes = self._cipher.block_size - (length % self._cipher.block_size)
        padding = struct.pack("!B", missing_bytes) * (missing_bytes + 1)

        if padding_cb:
            padding = padding_cb(bytearray(padding))

        fragment += padding

        if self._version <= tls.Version.TLS10:
            iv = self._iv

        else:
            # iv should be random but we want to have it reproducable
            iv = self._iv[:-4] + pdu.pack_uint32(self._seq_nbr)

        cipher = Cipher(self._cipher.algo(self._keys.enc), modes.CBC(iv))
        encryptor = cipher.encryptor()
        cipher_block = encryptor.update(fragment) + encryptor.finalize()
        if self._version <= tls.Version.TLS10:
            self._iv = cipher_block[-self._cipher.iv_len :]
            return cipher_block

        else:
            return iv + cipher_block

    def _append_mac(self, content_type, version, fragment, mac_cb=None, **kwargs):
        """Calculates the MAC and appends it to the fragment.

        Arguments:
            content_type (:obj:`tlsmate.constants.ContentType`): The content type.
            version (:obj:`tlsmate.constants.Version`): The version of the
                record layer.
            fragment: The message for which the MAC is calculated.

        Returns:
            bytes: The fragment appended with the MAC.
        """

        # TODO: move mac calculation to kdf object

        if version is tls.Version.SSL30:
            pad_len = 40 if self._mac.hash_algo.name == "sha1" else 48
            mac = hashes.Hash(self._mac.hash_algo())
            mac.update(
                self._keys.mac
                + b"\x36" * pad_len
                + pdu.pack_uint64(self._seq_nbr)
                + pdu.pack_uint8(content_type.value)
                + pdu.pack_uint16(len(fragment))
                + fragment
            )
            mac2 = hashes.Hash(self._mac.hash_algo())
            mac2.update(self._keys.mac + b"\x5c" * pad_len + mac.finalize())
            mac_bytes = mac2.finalize()

        else:
            mac_input = (
                pdu.pack_uint64(self._seq_nbr)
                + pdu.pack_uint8(content_type.value)
                + pdu.pack_uint16(version.value)
                + pdu.pack_uint16(len(fragment))
                + fragment
            )
            mac = hmac.HMAC(self._keys.mac, self._mac.hash_algo())
            mac.update(mac_input)
            mac_bytes = mac.finalize()

        if mac_cb:
            mac_bytes = mac_cb(bytearray(mac_bytes))

        return fragment + mac_bytes

    def _protect_block_cipher(self, rl_msg, data_cb=None, **kwargs):
        """Protects a fragment using a block cipher.

        The fragment is encrypted and authenticated with a MAC. Both modes,
        encrypt_then_mac and mac_then_encrypt are supported.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to protect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The protected record layer message.
        """

        fragment = rl_msg.fragment
        if data_cb:
            fragment = data_cb(bytearray(fragment))

        if self._enc_then_mac:
            fragment = self._encrypt_cbc(fragment, **kwargs)
            fragment = self._append_mac(
                rl_msg.content_type, rl_msg.version, fragment, **kwargs
            )

        else:
            fragment = self._append_mac(
                rl_msg.content_type, rl_msg.version, fragment, **kwargs
            )
            fragment = self._encrypt_cbc(fragment, **kwargs)

        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=rl_msg.content_type, version=rl_msg.version, fragment=fragment
        )

    def _protect_stream_cipher(self, rl_msg):
        """Protects a fragment using a stream cipher.

        The fragment is authenticated with a MAC and then encrypted.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to protect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The protected record layer message.
        """

        fragment = self._append_mac(
            rl_msg.content_type, rl_msg.version, rl_msg.fragment
        )
        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=rl_msg.content_type,
            version=rl_msg.version,
            fragment=self._cipher_object.update(fragment),
        )

    def _protect_chacha_cipher(self, rl_msg):
        """Protects a fragment using the CHACHA20_POLY1305 cipher.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to protect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The protected record layer message.
        """

        nonce_val = int.from_bytes(self._keys.iv, "big", signed=False) ^ self._seq_nbr
        nonce = nonce_val.to_bytes(self._cipher.iv_len, "big", signed=False)
        aad = (
            pdu.pack_uint64(self._seq_nbr)
            + pdu.pack_uint8(rl_msg.content_type.value)
            + pdu.pack_uint16(rl_msg.version.value)
            + pdu.pack_uint16(len(rl_msg.fragment))
        )
        chachapoly = aead.ChaCha20Poly1305(self._keys.enc)
        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=rl_msg.content_type,
            version=rl_msg.version,
            fragment=chachapoly.encrypt(nonce, rl_msg.fragment, aad),
        )

    def _protect_aead_cipher(self, rl_msg):
        """Protects a fragment using an AEAD cipher.

        This supports AES_GCM (128&256) as well as AES_CCM and AES_CCM_8.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to protect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The protected record layer message.
        """

        kwargs = {}
        if self._cipher.tag_length != 16:
            kwargs["tag_length"] = self._cipher.tag_length

        aes_aead = self._cipher.algo(self._keys.enc, **kwargs)
        nonce_explicit = pdu.pack_uint64(self._seq_nbr)
        nonce = self._keys.iv + nonce_explicit
        aad = (
            pdu.pack_uint64(self._seq_nbr)
            + pdu.pack_uint8(rl_msg.content_type.value)
            + pdu.pack_uint16(rl_msg.version.value)
            + pdu.pack_uint16(len(rl_msg.fragment))
        )
        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=rl_msg.content_type,
            version=rl_msg.version,
            fragment=nonce_explicit + aes_aead.encrypt(nonce, rl_msg.fragment, aad),
        )

    def _protect(self, rl_msg, **kwargs):
        """Protects a fragment.

        Supports stream ciphers, block cipher, AEAD ciphers and POLY20_CHACHA1305.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to protect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The protected record layer message.

        Raises:
            ValueError: if the cipher type is unknown.
        """

        if self._cipher.c_type == tls.CipherType.BLOCK:
            return self._protect_block_cipher(rl_msg, **kwargs)

        elif self._cipher.c_type == tls.CipherType.STREAM:
            return self._protect_stream_cipher(rl_msg)

        elif self._cipher.c_type == tls.CipherType.AEAD:
            if self._cipher.primitive == tls.CipherPrimitive.CHACHA:
                return self._protect_chacha_cipher(rl_msg)

            else:
                return self._protect_aead_cipher(rl_msg)

        else:
            raise ValueError("Unknown cipher type")

    def _tls13_protect(self, rl_msg):
        """Protects a fragment using the TLS1.3 specification.

        All TLS1.3 ciphers are supported.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to protect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The protected record layer message.
        """

        fragment = bytes(rl_msg.fragment) + pdu.pack_uint8(rl_msg.content_type.value)
        aad = (
            pdu.pack_uint8(tls.ContentType.APPLICATION_DATA.value)
            + pdu.pack_uint16(rl_msg.version.value)
            + pdu.pack_uint16(len(fragment) + self._cipher.tag_length)
        )
        nonce_val = int.from_bytes(self._keys.iv, "big", signed=False) ^ self._seq_nbr
        nonce = nonce_val.to_bytes(self._cipher.iv_len, "big", signed=False)
        kwargs = {}
        if self._cipher.tag_length == 8:
            kwargs["tag_length"] = 8

        cipher = self._cipher_object(self._keys.enc, **kwargs)
        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=tls.ContentType.APPLICATION_DATA,
            version=rl_msg.version,
            fragment=cipher.encrypt(nonce, fragment, aad),
        )

    def protect_msg(
        self, rl_msg: structs.RecordLayerMsg, **kwargs: Any
    ) -> structs.RecordLayerMsg:
        """Protects a fragment.

        Arguments:
            rl_msg: The record layer message to protect.
            kwargs: additional parameters which can be used to control CBC
                padding oracle related behavior

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The protected record layer message.
        """

        if self._version is tls.Version.TLS13:
            return self._tls13_protect(rl_msg)

        else:
            # Skip compression, we don't want to support it.
            return self._protect(rl_msg, **kwargs)

    def _verify_mac(self, content_type, version, fragment, mac_cb=None, **kwargs):
        """Verifies the MAC of a fragment.

        Arguments:
            content_type (:obj:`tlsmate.constants.ContentType`): The content type.
            version (:obj:`tlsmate.constants.Version`): The version of the
                record layer.
            fragment: The message for which the MAC is verified.

        Returns:
            bytes: The fragment appended with the MAC.

        Raises:
            :obj:`tlsmate.tls.ServerMalfunction`: If the MAC is incorrect.
        """

        if len(fragment) < self._mac.mac_len:
            raise tls.ServerMalfunction(tls.ServerIssue.RECORD_TOO_SHORT)

        msg_len = len(fragment) - self._mac.mac_len
        mac_received = fragment[msg_len:]
        if mac_cb:
            mac_received = mac_cb(bytearray(mac_received))

        msg = fragment[:msg_len]

        if version is tls.Version.SSL30:
            pad_len = 40 if self._mac.hash_algo.name == "sha1" else 48
            mac = hashes.Hash(self._mac.hash_algo())
            mac.update(
                self._keys.mac
                + b"\x36" * pad_len
                + pdu.pack_uint64(self._seq_nbr)
                + pdu.pack_uint8(content_type.value)
                + pdu.pack_uint16(msg_len)
                + msg
            )
            mac2 = hashes.Hash(self._mac.hash_algo())
            mac2.update(self._keys.mac + b"\x5c" * pad_len + mac.finalize())
            mac_calculated = mac2.finalize()

        else:
            mac_input = (
                pdu.pack_uint64(self._seq_nbr)
                + pdu.pack_uint8(content_type.value)
                + pdu.pack_uint16(version.value)
                + pdu.pack_uint16(msg_len)
                + msg
            )
            mac = hmac.HMAC(self._keys.mac, self._mac.hash_algo())
            mac.update(mac_input)
            mac_calculated = mac.finalize()

        if mac_calculated != mac_received:
            raise tls.ServerMalfunction(tls.ServerIssue.RECORD_MAC_INVALID)

        return msg

    def _decode_cbc(self, fragment, padding_cb=None, **kwargs):
        """Decodes a fragment using a block cipher in CBC mode.

        Arguments:
            fragment (bytes): The fragment to decode.

        Returns:
            bytes: the decoded fragment.

        Raises:
            :obj:`tlsmate.tls.ServerMalfunction`: If padding errors are detected.
        """

        if self._version <= tls.Version.TLS10:
            iv = self._iv
            cipher_text = fragment
            self._iv = fragment[-self._cipher.iv_len :]

        else:
            iv = fragment[: self._cipher.iv_len]
            cipher_text = fragment[self._cipher.iv_len :]

        cipher = Cipher(self._cipher.algo(self._keys.enc), modes.CBC(iv))
        decryptor = cipher.decryptor()
        plain_text = decryptor.update(cipher_text) + decryptor.finalize()

        # padding
        pad = plain_text[-1]
        pad_start = len(plain_text) - pad - 1
        if pad_start < 0:
            raise tls.ServerMalfunction(tls.ServerIssue.RECORD_WRONG_PADDING_LENGTH)

        padding = plain_text[pad_start:]
        if padding_cb:
            padding = padding_cb(bytearray(padding))

        plain_text = plain_text[:pad_start]

        if self._version is not tls.Version.SSL30:
            if (struct.pack("!B", pad) * (pad + 1)) != padding:
                raise tls.ServerMalfunction(tls.ServerIssue.RECORD_WRONG_PADDING_BYTES)

        return plain_text

    def _unprotect_block_cipher(self, rl_msg, data_cb=None, **kwargs):
        """Unprotects a record layer message (block cipher).

        The message is decrypted and the authentication is verified. Both modes,
        encrypt_then_mac and mac_then_encrypt are supported.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to unprotect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The unprotected record layer message (plain text).
        """

        if self._enc_then_mac:
            fragment = self._verify_mac(
                rl_msg.content_type, rl_msg.version, rl_msg.fragment, **kwargs,
            )
            plain_text = self._decode_cbc(fragment, **kwargs)

        else:
            fragment = self._decode_cbc(rl_msg.fragment, **kwargs)
            plain_text = self._verify_mac(
                rl_msg.content_type, rl_msg.version, fragment, **kwargs
            )

        if data_cb:
            plain_text = data_cb(bytearray(plain_text))

        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=rl_msg.content_type,
            version=rl_msg.version,
            fragment=plain_text,
        )

    def _unprotect_stream_cipher(self, rl_msg, **kwargs):
        """Unprotects a record layer message (stream cipher).

        The message is decrypted and the authentication is verified.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to unprotect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The unprotected record layer message (plain text).
        """

        fragment = self._cipher_object.update(rl_msg.fragment)
        clear_text = self._verify_mac(
            rl_msg.content_type, rl_msg.version, fragment, **kwargs
        )
        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=rl_msg.content_type,
            version=rl_msg.version,
            fragment=clear_text,
        )

    def _unprotect_chacha_cipher(self, rl_msg):
        """Unprotects a record layer message (CHACHA20_POLY1305 cipher).

        The message is decrypted and the authentication is verified.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to unprotect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The unprotected record layer message (plain text).
        """

        nonce_val = int.from_bytes(self._keys.iv, "big", signed=False) ^ self._seq_nbr
        nonce = nonce_val.to_bytes(self._cipher.iv_len, "big", signed=False)
        aad = (
            pdu.pack_uint64(self._seq_nbr)
            + pdu.pack_uint8(rl_msg.content_type.value)
            + pdu.pack_uint16(rl_msg.version.value)
            + pdu.pack_uint16(len(rl_msg.fragment) - self._cipher.tag_length)
        )
        chachapoly = aead.ChaCha20Poly1305(self._keys.enc)
        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=rl_msg.content_type,
            version=rl_msg.version,
            fragment=chachapoly.decrypt(nonce, rl_msg.fragment, aad),
        )

    def _unprotect_aead_cipher(self, rl_msg):
        """Unprotects a record layer message (AEAD cipher).

        The message is decrypted and the authentication is verified.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to unprotect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The unprotected record layer message (plain text).
        """

        nonce_explicit = rl_msg.fragment[:8]
        cipher_text = rl_msg.fragment[8:]
        aad = (
            pdu.pack_uint64(self._seq_nbr)
            + pdu.pack_uint8(rl_msg.content_type.value)
            + pdu.pack_uint16(rl_msg.version.value)
            + pdu.pack_uint16(len(cipher_text) - self._cipher.tag_length)
        )
        nonce = bytes(self._keys.iv + nonce_explicit)
        kwargs = {}
        if self._cipher.tag_length != 16:
            kwargs["tag_length"] = self._cipher.tag_length

        aes_aead = self._cipher.algo(self._keys.enc, **kwargs)
        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=rl_msg.content_type,
            version=rl_msg.version,
            fragment=aes_aead.decrypt(nonce, cipher_text, aad),
        )

    def _unprotect(self, rl_msg, **kwargs):
        """Unprotects a record layer message (all ciphers, but not for TLS1.3).

        The message is decrypted and the authentication is verified.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to unprotect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The unprotected record layer message (plain text).
        """

        if self._cipher.c_type == tls.CipherType.BLOCK:
            return self._unprotect_block_cipher(rl_msg, **kwargs)

        elif self._cipher.c_type == tls.CipherType.STREAM:
            return self._unprotect_stream_cipher(rl_msg, **kwargs)

        elif self._cipher.c_type == tls.CipherType.AEAD:
            if self._cipher.primitive == tls.CipherPrimitive.CHACHA:
                return self._unprotect_chacha_cipher(rl_msg)

            else:
                return self._unprotect_aead_cipher(rl_msg)

        else:
            raise ValueError("Unknown cipher type")

    def _tls13_unprotect(self, rl_msg):
        """Unprotects a record layer message (TLS1.3).

        The message is decrypted and the authentication is verified.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to unprotect.

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The unprotected record layer message (plain text).

        Raises:
            :obj:`tlsmate.tls.ServerMalfunction`: If padding errors are detected.
        """

        aad = (
            pdu.pack_uint8(rl_msg.content_type.value)
            + pdu.pack_uint16(rl_msg.version.value)
            + pdu.pack_uint16(len(rl_msg.fragment))
        )
        nonce_val = int.from_bytes(self._keys.iv, "big", signed=False) ^ self._seq_nbr
        nonce = nonce_val.to_bytes(self._cipher.iv_len, "big", signed=False)
        kwargs = {}
        if self._cipher.tag_length == 8:
            kwargs["tag_length"] = 8

        cipher = self._cipher_object(self._keys.enc, **kwargs)
        decoded = cipher.decrypt(nonce, rl_msg.fragment, aad)
        # find idx of last non-zero octet
        idx = len(decoded) - 1
        while idx >= 0:
            if decoded[idx] != 0:
                break

            idx -= 1

        if idx < 0:
            raise tls.ServerMalfunction(tls.ServerIssue.RECORD_WRONG_PADDING_LENGTH)

        self._seq_nbr += 1
        return structs.RecordLayerMsg(
            content_type=tls.ContentType.val2enum(decoded[idx], alert_on_failure=True),
            version=rl_msg.version,
            fragment=decoded[:idx],
        )

    def unprotect_msg(
        self, rl_msg: structs.RecordLayerMsg, **kwargs
    ) -> structs.RecordLayerMsg:
        """Unprotects a record layer message.

        The message is decrypted and the authentication is verified.

        Arguments:
            rl_msg (:obj:`tlsmate.structs.RecordLayerMsg`): The record layer
                message to unprotect.

            kwargs (dict): additional parameters which can be used to control CBC
                padding oracle related behavior

        Returns:
            :obj:`tlsmate.structs.RecordLayerMsg`:
            The unprotected record layer message (plain text).
        """

        if self._version is tls.Version.TLS13:
            if rl_msg.content_type is tls.ContentType.CHANGE_CIPHER_SPEC:
                return rl_msg

            return self._tls13_unprotect(rl_msg)

        else:
            # We do not support compression.
            return self._unprotect(rl_msg, **kwargs)



# -*- coding: utf-8 -*-

# import basic stuff

# import own stuff
import tlsmate.ext as ext
import tlsmate.tls as tls

import pytest

# import other stuff


def test_ServerNameIndication():
    data = bytes.fromhex("00 00 00 0e 00 0c 00 00 09 6c 6f 63 61 6c 68 6f 73 74")
    sni, offset = ext.Extension.deserialize(data, 0)
    assert offset == len(data)
    assert sni.host_name == "localhost"
    assert sni.serialize(None) == data
    sni = ext.ExtServerNameIndication(host_name="localhost")
    assert sni.serialize(None) == data


def test_sni_empty():
    data = bytes.fromhex("00 00 00 00")
    sni, offset = ext.Extension.deserialize(data, 0)
    assert sni.host_name is None


def test_sni_wrong_length():
    data = bytes.fromhex("00 00 00 0e 00 0d 00 00 09 6c 6f 63 61 6c 68 6f 73 74")
    with pytest.raises(tls.ServerMalfunction, match="extension length incorrect"):
        sni, offset = ext.Extension.deserialize(data, 0)


def test_sni_wrong_name_type():
    data = bytes.fromhex("00 00 00 0e 00 0c 01 00 09 6c 6f 63 61 6c 68 6f 73 74")
    with pytest.raises(tls.ServerMalfunction, match="host_name not present"):
        sni, offset = ext.Extension.deserialize(data, 0)


def test_ems_wrong_length():
    data = bytes.fromhex("00 17 00 01 00")
    with pytest.raises(tls.ServerMalfunction, match="extension length incorrect"):
        ems, offset = ext.Extension.deserialize(data, 0)


def test_etm_wrong_length():
    data = bytes.fromhex("00 16 00 01 00")
    with pytest.raises(tls.ServerMalfunction, match="extension length incorrect"):
        ems, offset = ext.Extension.deserialize(data, 0)


def test_ec_point_format():
    epf = ext.ExtEcPointFormats(ec_point_formats=[tls.EcPointFormat.UNCOMPRESSED, 5])
    assert epf.serialize(None) == bytes.fromhex("00 0b 00 03 02 00 05")


def test_ec_point_format_wrong_length():
    data = bytes.fromhex("00 0b 00 03 03 00 05")
    with pytest.raises(tls.ServerMalfunction, match="extension length incorrect"):
        epf, offset = ext.Extension.deserialize(data, 0)


def test_supported_groups_unknown_value():
    data = bytes.fromhex("00 0a 00 0a 00 08 00 1d 00 17 00 18 00 aa")
    groups, offset = ext.Extension.deserialize(data, 0)
    assert offset == len(data)
    assert type(groups) is ext.ExtSupportedGroups
    assert groups.supported_groups == [
        tls.SupportedGroups.X25519,
        tls.SupportedGroups.SECP256R1,
        tls.SupportedGroups.SECP384R1,
        0x00AA,
    ]


def test_sig_algo():
    extension = ext.ExtSignatureAlgorithms(
        signature_algorithms=[tls.SignatureScheme.ECDSA_SHA1, (0xAA, 0xBB)]
    )
    data = extension.serialize(None)
    assert data == bytes.fromhex("00 0d 00 06 00 04 02 03 aa bb")
    sig_algo, offset = ext.Extension.deserialize(data, 0)
    assert offset == len(data)
    assert type(sig_algo) is ext.ExtSignatureAlgorithms
    assert sig_algo.signature_algorithms == [tls.SignatureScheme.ECDSA_SHA1, 0xAABB]


def test_heartbeat():
    extension = ext.ExtHeartbeat(heartbeat_mode=0x11)
    assert extension.serialize(None) == bytes.fromhex("00 0f 00 01 11")


def test_cert_authorities():
    data = bytes.fromhex("00 2f 00 0e 00 0c 00 03 aa bb cc 00 05 11 22 33 44 55")
    ca, offset = ext.Extension.deserialize(data, 0)
    assert offset == len(data)
    assert type(ca) is ext.ExtCertificateAuthorities
    assert len(ca.authorities) == 2
    assert ca.authorities[0] == bytes.fromhex("aa bb cc")
    assert ca.authorities[1] == bytes.fromhex("11 22 33 44 55")


def test_status_request():
    extension = ext.ExtStatusRequest(responder_ids=[b"\4\5"], extensions=[b"\1"])
    data = bytes.fromhex("00 05 00 0c 01  00 04 00 02 04 05  00 03 00 01 01")
    assert extension.serialize(None) == data


def test_status_request_v2():
    extension = ext.ExtStatusRequestV2(responder_ids=[b"\4\5"], extensions=b"\1")
    data = bytes.fromhex("00 11 00 0e 00 0c 02 00 09 00 04 00 02 04 05  00 01 01")
    assert extension.serialize(None) == data


def test_status_request_v2_deserialize():
    data = bytes.fromhex("00 11 00 09 02 00 00 05 01 02 03 04 05")
    status, offset = ext.Extension.deserialize(data, 0)
    assert offset == len(data)
    assert type(status) is ext.ExtStatusRequestV2
    assert status.status_type is tls.StatusType.OCSP_MULTI
    assert status.ocsp_response == bytes.fromhex("01 02 03 04 05")


def test_unknown_ext():
    data = bytes.fromhex("00 ab 00 0e 00 0c 00 00 09 6c 6f 63 61 6c 68 6f 73 74")
    extension, offset = ext.Extension.deserialize(data, 0)
    assert offset == len(data)
    assert extension.extension_id is tls.Extension.UNKNOW_EXTENSION
    assert extension.id == 0x00AB
    assert extension.bytes == data[4:]
    assert extension.serialize(None) == data
    ext2 = ext.ExtUnknownExtension(
        id=0x00AB, bytes=bytes.fromhex("00 0c 00 00 09 6c 6f 63 61 6c 68 6f 73 74")
    )
    assert ext2.serialize(None) == data



# -*- coding: utf-8 -*-
"""Define common fixtures
"""
import pytest
import pathlib
import pem
import os
import datetime
import subprocess
import time

import tlsmate.tlsmate as tm


@pytest.fixture
def proxy():
    port = os.environ.get("TLSMATE_PROXY_PORT", 8801)
    cmd = f"ncat -l --proxy-type http localhost {port}"
    proc = subprocess.Popen(cmd.split())
    time.sleep(1)
    yield f"http://localhost:{port}"
    proc.kill()


@pytest.fixture
def valid_time():
    return datetime.datetime(2031, 6, 1)


@pytest.fixture
def style_file():
    return pathlib.Path(__file__).parent.resolve() / "../tlsmate/styles/default.yaml"


@pytest.fixture
def fixturefiles_dir():
    return pathlib.Path(__file__).parent.resolve() / "fixturefiles"


@pytest.fixture
def tlsmate_empty_ini(fixturefiles_dir):
    return fixturefiles_dir / "tlsmate_empty.ini"


@pytest.fixture
def guballa_de_pem(fixturefiles_dir):
    return pem.parse_file(fixturefiles_dir / "guballa.de.pem")[0]


@pytest.fixture
def quo_vadis_root_ca3(fixturefiles_dir):
    return pem.parse_file(fixturefiles_dir / "QuoVadis_Root_CA_3.pem")[0]


@pytest.fixture
def digi_cert_global_root_ca_file(fixturefiles_dir):
    return fixturefiles_dir / "DigiCertGlobalRootCA.pem"


@pytest.fixture
def ca_dir(fixturefiles_dir):
    return fixturefiles_dir / "../ca"


@pytest.fixture
def trust_store_file(ca_dir):
    return ca_dir / "certs/root-certificates.pem"


@pytest.fixture
def ca_rsa_crl_file(ca_dir):
    return ca_dir / "crl/ca-rsa.crl.pem"


@pytest.fixture
def server_rsa_cert_file(ca_dir):
    return ca_dir / "certs/server-rsa.pem"


@pytest.fixture
def server_rsa_key_file(ca_dir):
    return ca_dir / "private/server-rsa.key"


@pytest.fixture
def server_rsa_chain_file(ca_dir):
    return ca_dir / "chains/server-rsa.chn"


@pytest.fixture
def tlsmate(ca_dir, trust_store_file, digi_cert_global_root_ca_file):
    mate = tm.TlsMate()
    mate.trust_store.set_ca_files([trust_store_file, digi_cert_global_root_ca_file])
    return mate


@pytest.fixture
def server_expired_rsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/server-expired-rsa.pem")[0]


@pytest.fixture
def server_no_ids_rsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/server-no-ids-rsa.pem")[0]


@pytest.fixture
def server_revoked_rsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/server-revoked-rsa.pem")[0]


@pytest.fixture
def server_rsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/server-rsa.pem")[0]


@pytest.fixture
def server_dsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/server-dsa.pem")[0]


@pytest.fixture
def server_ed25519_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/server-ed25519.pem")[0]


@pytest.fixture
def server_ed448_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/server-ed448.pem")[0]


@pytest.fixture
def ca_rsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/ca-rsa.pem")[0]


@pytest.fixture
def ca_rsa_key(ca_dir):
    return pem.parse_file(ca_dir / "private/ca-rsa.key")[0]


@pytest.fixture
def ca_2nd_rsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/ca-2nd-rsa.pem")[0]


@pytest.fixture
def ca_ecdsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/ca-ecdsa.pem")[0]


@pytest.fixture
def root_rsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/root-rsa.pem")[0]


@pytest.fixture
def root_rsa_key(ca_dir):
    return pem.parse_file(ca_dir / "private/root-rsa.key")[0]


@pytest.fixture
def root_ecdsa_cert(ca_dir):
    return pem.parse_file(ca_dir / "certs/root-ecdsa.pem")[0]


@pytest.fixture
def client_rsa_key_filename(ca_dir):
    return str(ca_dir / "private/client-rsa.key")


@pytest.fixture
def client_rsa_chain_filename(ca_dir):
    return str(ca_dir / "chains/client-rsa.chn")


@pytest.fixture
def server_profile(fixturefiles_dir):
    return fixturefiles_dir / "server_profile.yaml"


@pytest.fixture
def text_server_profile(fixturefiles_dir):
    return fixturefiles_dir / "text_server_profile.yaml"


@pytest.fixture
def server_profile_base_vuln(fixturefiles_dir):
    return fixturefiles_dir / "server_profile_base_vulnerabilities.yaml"


@pytest.fixture
def server_profile_no_compr(fixturefiles_dir):
    return fixturefiles_dir / "server_profile_no_compr.yaml"


@pytest.fixture
def server_profile_no_features(fixturefiles_dir):
    return fixturefiles_dir / "server_profile_no_features.yaml"


@pytest.fixture
def server_profile_logjam_common(fixturefiles_dir):
    return fixturefiles_dir / "server_profile_logjam_common.yaml"


@pytest.fixture
def server_profile_logjam_cust(fixturefiles_dir):
    return fixturefiles_dir / "server_profile_logjam_customized.yaml"


@pytest.fixture
def server_profile_no_dh_group(fixturefiles_dir):
    return fixturefiles_dir / "server_profile_no_dh_group.yaml"


@pytest.fixture
def server_profile_no_dh(fixturefiles_dir):
    return fixturefiles_dir / "server_profile_no_dh.yaml"


@pytest.fixture
def full_server_profile(fixturefiles_dir):
    return fixturefiles_dir / "full_server_profile.yaml"



# -*- coding: utf-8 -*-
"""Implement unit tests for the module recorder.
"""
import pytest
import dns.exception
import dns.resolver
import tlsmate.tls as tls
import tlsmate.structs as structs
import tlsmate.resolver as resolver


@pytest.fixture(autouse=True)
def clear_cache():
    resolver._resolved = {}


def test_determine_l4_addr():
    l4_endpoint = resolver.determine_l4_addr("google.com", 443)
    assert l4_endpoint.host == "google.com"
    assert l4_endpoint.port == 443
    assert l4_endpoint.host_type is tls.HostType.HOST

    l4_endpoint = resolver.determine_l4_addr("123.0.1.12", 443)
    assert l4_endpoint.host == "123.0.1.12"
    assert l4_endpoint.port == 443
    assert l4_endpoint.host_type is tls.HostType.IPV4

    l4_endpoint = resolver.determine_l4_addr("fe08::23", 443)
    assert l4_endpoint.host == "fe08::23"
    assert l4_endpoint.port == 443
    assert l4_endpoint.host_type is tls.HostType.IPV6


def test_resolve_hostname():
    resolved = resolver.resolve_hostname("example.com")
    assert resolved.ipv4_addresses == ["93.184.216.34"]
    assert resolved.ipv6_addresses == ["2606:2800:220:1:248:1893:25c8:1946"]


def test_resolve_hostname_proxy(proxy):
    resolved = resolver.resolve_hostname("example.com", proxy=proxy)
    assert resolved.ipv4_addresses == ["93.184.216.34"]
    assert resolved.ipv6_addresses == ["2606:2800:220:1:248:1893:25c8:1946"]


def test_resolve_hostname_proxy_timeout(monkeypatch):
    def resolve(self, host_name, rd_type):
        raise dns.exception.Timeout

    monkeypatch.setattr(dns.resolver.Resolver, "resolve", resolve)

    resolved = resolver.resolve_hostname("example.com", proxy="http://localhost:8889")
    assert resolved.ipv4_addresses == []
    assert resolved.ipv6_addresses == []


def test_resolve_hostname_proxy_no_answer(monkeypatch):
    def resolve(self, host_name, rd_type):
        raise dns.resolver.NoAnswer

    monkeypatch.setattr(dns.resolver.Resolver, "resolve", resolve)

    resolved = resolver.resolve_hostname("example.com", proxy="http://localhost:8889")
    assert resolved.ipv4_addresses == []
    assert resolved.ipv6_addresses == []


def test_get_ip_endpoint_resolved():
    endpoint = structs.TransportEndpoint(
        host="123.123.123", port=443, host_type=tls.HostType.IPV4
    )
    resolved = resolver.get_ip_endpoint(endpoint)
    assert endpoint is resolved


def test_get_ip_endpoint_host_ipv4():
    endpoint = structs.TransportEndpoint(
        host="example.com", port=443, host_type=tls.HostType.HOST
    )
    resolved = resolver.get_ip_endpoint(endpoint)
    assert resolved.host == "93.184.216.34"
    assert resolved.port == 443
    assert resolved.host_type is tls.HostType.IPV4


def test_get_ip_endpoint_host_ipv6():
    endpoint = structs.TransportEndpoint(
        host="ipv6.test-ipv6.com", port=443, host_type=tls.HostType.HOST
    )
    resolved = resolver.get_ip_endpoint(endpoint)
    assert resolved.host == "2001:470:1:18::115"
    assert resolved.port == 443
    assert resolved.host_type is tls.HostType.IPV6


def test_get_ip_endpoint_ipv6_preference():
    endpoint = structs.TransportEndpoint(
        host="example.com", port=443, host_type=tls.HostType.HOST
    )
    resolved = resolver.get_ip_endpoint(endpoint, ipv6_preference=True)
    assert resolved.host == "2606:2800:220:1:248:1893:25c8:1946"
    assert resolved.port == 443
    assert resolved.host_type is tls.HostType.IPV6



# -*- coding: utf-8 -*-
"""Module scanning for extended-master-secret support
"""
# import basic stuff

# import own stuff
import tlsmate.plugin as plg
import tlsmate.tls as tls

# import other stuff


class ScanExtendedMasterSecret(plg.Worker):
    name = "master_secret"
    descr = "scan for extension extended_master_secret support"
    prio = 30

    def run(self):
        state = tls.ScanState.UNDETERMINED
        versions = [tls.Version.TLS10, tls.Version.TLS11, tls.Version.TLS12]
        prof_values = self.server_profile.get_profile_values(versions, full_hs=True)
        if not prof_values.versions:
            state = tls.ScanState.NA
        else:
            self.client.init_profile(profile_values=prof_values)
            self.client.profile.support_extended_master_secret = True
            with self.client.create_connection() as conn:
                conn.handshake()
            if conn.handshake_completed:
                if conn.msg.server_hello.get_extension(
                    tls.Extension.EXTENDED_MASTER_SECRET
                ):
                    state = tls.ScanState.TRUE
                else:
                    state = tls.ScanState.FALSE

        self.server_profile.allocate_features()
        self.server_profile.features.extended_master_secret = state



# -*- coding: utf-8 -*-
"""Module for logging to a key log file

Reference: https://developer.mozilla.org/en-US/docs/Mozilla/Projects/NSS/Key_Log_Format
"""
# import basic stuff
from typing import Optional, TextIO

# import own stuff

# import other stuff


class KeyLogger(object):
    """Class to log to a key log file
    """

    _fd: Optional[TextIO] = None

    @classmethod
    def open_file(cls, file_name: str) -> None:
        """Opens the key logging file in "append" mode.

        Arguments:
            file_name (str): the file name of the key logging file
        """

        cls._fd = open(file_name, "a")

    @classmethod
    def close(cls) -> None:
        """Closes the key logging file.
        """

        assert cls._fd
        cls._fd.close()

    @classmethod
    def _log(cls, log_type: str, client_random: bytes, secret: bytes) -> None:
        """Adds a log to the key logging file.

        Arguments:
            log_type: the type of the log, first element of the log entry
            client_random: the random value from the ClientHello
            secret: the secret to log
        """

        if cls._fd:
            cls._fd.write(f"{log_type} {client_random.hex()} {secret.hex()}\n")
            cls._fd.flush()

    @classmethod
    def master_secret(cls, client_random: bytes, secret: bytes) -> None:
        """Generates a CLIENT_RANDOM log.

        Arguments:
            client_random: the random value from the ClientHello
            secret: the secret to log
        """

        cls._log("CLIENT_RANDOM", client_random, secret)

    @classmethod
    def client_early_tr_secret(cls, client_random: bytes, secret: bytes) -> None:
        """Generates a CLIENT_EARLY_TRAFFIC_SECRET log.

        Arguments:
            client_random: the random value from the ClientHello
            secret: the client early traffic secret
        """

        cls._log("CLIENT_EARLY_TRAFFIC_SECRET", client_random, secret)

    @classmethod
    def client_hs_tr_secret(cls, client_random: bytes, secret: bytes) -> None:
        """Generates a CLIENT_HANDSHAKE_TRAFFIC_SECRET log.

        Arguments:
            client_random: the random value from the ClientHello
            secret: the client handshake traffic secret
        """

        cls._log("CLIENT_HANDSHAKE_TRAFFIC_SECRET", client_random, secret)

    @classmethod
    def server_hs_tr_secret(cls, client_random: bytes, secret: bytes) -> None:
        """Generates a SERVER_HANDSHAKE_TRAFFIC_SECRET log.

        Arguments:
            client_random: the random value from the ClientHello
            secret: the server handshake traffic secret
        """

        cls._log("SERVER_HANDSHAKE_TRAFFIC_SECRET", client_random, secret)

    @classmethod
    def client_tr_secret_0(cls, client_random: bytes, secret: bytes) -> None:
        """Generates a CLIENT_TRAFFIC_SECRET_0 log.

        Arguments:
            client_random: the random value from the ClientHello
            secret: the client traffic secret
        """

        cls._log("CLIENT_TRAFFIC_SECRET_0", client_random, secret)

    @classmethod
    def server_tr_secret_0(cls, client_random: bytes, secret: bytes) -> None:
        """Generates a SERVER_TRAFFIC_SECRET_0 log.

        Arguments:
            client_random: the random value from the ClientHello
            secret: the server traffic secret
        """

        cls._log("SERVER_TRAFFIC_SECRET_0", client_random, secret)



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
import tests.cipher_suite_tester as tester
import tlsmate.tls as tls
import tlsmate.tlssuite as tlssuite


class TestCase(tester.CipherSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the CipherSuiteTester class.
    """

    path = pathlib.Path(__file__)

    cipher_suite = tls.CipherSuite.TLS_DH_ANON_WITH_AES_128_CBC_SHA
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa -- -www -cipher ALL"
    )
    library = tlssuite.TlsLibrary.openssl1_0_2

    # Uncomment the line below if you do not want to use the default version and
    # adapt it to your needs.
    version = tls.Version.TLS10


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module defining various structures
"""
# import basic stuff
from typing import NamedTuple, Any, Optional, Type

# import own stuff
import tlsmate.kdf as kdf
import tlsmate.tls as tls

# import other stuff
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives._asymmetric import AsymmetricPadding


class SessionStateId(NamedTuple):
    """Set of items to store a session id in the client.
    """

    session_id: bytes
    cipher_suite: tls.CipherSuite
    version: tls.Version
    master_secret: bytes


class SessionStateTicket(NamedTuple):
    """Set of items to store a session ticket in the client.
    """

    ticket: bytes
    lifetime: int
    cipher_suite: tls.CipherSuite
    version: tls.Version
    master_secret: bytes


class EarlyData(NamedTuple):
    """Set of items required to determine the early data traffic keys
    """

    kdf: kdf.Kdf
    early_secret: bytes
    mac_len: int
    binders_bytes: Optional[bytes] = None


class Cipher(NamedTuple):
    """Set of properties describing a cipher.
    """

    primitive: Optional[tls.CipherPrimitive] = None
    algo: Optional[Any] = None
    c_type: Optional[tls.CipherType] = None
    key_len: Optional[int] = None
    block_size: Optional[int] = None
    iv_len: Optional[int] = None
    tag_length: Optional[int] = None
    cipher_supported: bool = False


class Mac(NamedTuple):
    """Set of properties describing a MAC.
    """

    hash_algo: Type[hashes.HashAlgorithm]
    mac_len: Optional[int]
    key_len: Optional[int]
    hmac_algo: Optional[Type[hashes.HashAlgorithm]]


class Psk(NamedTuple):
    """Set of items to store a TLS1.3 pre-shared key in the client.
    """

    psk: bytearray
    lifetime: int
    age_add: int
    ticket: bytes
    timestamp: float
    cipher_suite: tls.CipherSuite
    version: tls.Version
    hmac: Mac


class SymmetricKeys(NamedTuple):
    """Set of keys
    """

    mac: bytes
    enc: bytes
    iv: bytes


class StateUpdateParams(NamedTuple):
    """Set of properties used to update the record layer state.
    """

    cipher: Cipher
    mac: Mac
    keys: SymmetricKeys
    compr: tls.CompressionMethod
    enc_then_mac: bool
    version: tls.Version
    is_write_state: bool


class CipherSuite(NamedTuple):
    """Set of properties for a cipher suite.
    """

    key_ex: tls.KeyExchangeAlgorithm
    cipher: tls.SymmetricCipher
    mac: tls.HashPrimitive


class RecordLayerMsg(NamedTuple):
    """Set of properties describing a record layer message.
    """

    content_type: tls.ContentType
    version: tls.Version
    fragment: bytes


class UpperLayerMsg(NamedTuple):
    """Set of properties describing an upper layer message.

    "Upper Layer" comprises handshake, ccs and alert protocols.
    """

    content_type: tls.ContentType
    msg_type: tls.HandshakeType  # can be CCSType as well
    msg: bytes


class KeyExchange(NamedTuple):
    """Set of properties describing a key exchange method.
    """

    key_ex_type: Optional[tls.KeyExchangeType] = None
    key_auth: Optional[tls.KeyAuthentication] = None
    key_ex_supported: bool = False
    default_sig_scheme: Optional[tls.SignatureScheme] = None


class KeyShareEntry(NamedTuple):
    """Set of properties describing a key share entry.
    """

    group: tls.SupportedGroups
    key_exchange: Optional[bytes]


class CipherSuiteDetails(NamedTuple):
    """Structure which provides details for a cipher suite.
    """

    cipher_suite: tls.CipherSuite
    full_hs: bool = False
    key_exchange_supported: bool = False
    key_algo: Optional[tls.KeyExchangeAlgorithm] = None
    key_algo_struct: Optional[KeyExchange] = None
    cipher: Optional[tls.SymmetricCipher] = None
    cipher_struct: Optional[Cipher] = None
    mac: Optional[tls.HashPrimitive] = None
    mac_struct: Optional[Mac] = None


class ProfileValues(NamedTuple):
    """Structure for the most relevant parameters for client hello
    """

    versions: list = []
    cipher_suites: list = []
    supported_groups: list = []
    signature_algorithms: list = []
    key_shares: list = []
    """Only applicable for TLS1.3"""


class CertSigAlgo(NamedTuple):
    """Structure for certificate signature algorithms
    """

    algo: Optional[Type[hashes.HashAlgorithm]] = None
    padd: Optional[Type[AsymmetricPadding]] = None


class TransportEndpoint(NamedTuple):
    """Structure for host, type and port
    """

    host: str
    port: int
    host_type: tls.HostType


class ResolvedHost(NamedTuple):
    """Result of a DNS resolution
    """

    ipv4_addresses: list
    ipv6_addresses: list


class ConfigItem(NamedTuple):
    """A configuration setting
    """

    name: str
    default: Any = None
    type: Any = str


class Malfunction(NamedTuple):
    """Structure for server malfunction
    """

    issue: tls.ServerIssue
    message: Optional[tls.HandshakeType] = None
    extension: Optional[tls.Extension] = None



# -*- coding: utf-8 -*-
"""Module containing the CLI implementation
"""
# import basic stuff
import importlib
import argparse
import logging

# import own stuff
import tlsmate.config as conf
import tlsmate.plugin as plg
import tlsmate.tlsmate as tm
import tlsmate.utils as utils

# import other stuff


class _PreParser(plg.Plugin):
    plugins = [plg.ArgConfig, plg.ArgPlugin, plg.ArgLogging]


def build_parser() -> argparse.ArgumentParser:
    """Creates the parser object

    Returns:
        the parser object as created with argparse
    """

    return plg.BaseCommand.create_parser()


def _load_plugins(config: conf.Configuration) -> None:

    pre_parser = argparse.ArgumentParser(add_help=False)
    _PreParser.extend_parser(pre_parser, None)
    pre_args, _ = pre_parser.parse_known_args()

    # logging should be setup as early as possible
    utils.set_logging_level(pre_args.logging)

    if pre_args.plugin:
        plugins = pre_args.plugin

    else:
        plugins = config.get_from_external(pre_args.config_file, "plugin")

    if plugins:
        for plugin in plugins:
            if plugin.startswith("tlsmate_"):
                try:
                    importlib.import_module(plugin)
                    logging.debug(f"Plugin module {plugin} successfully loaded")

                except ModuleNotFoundError:
                    pass


def main() -> None:
    """The entry point for the command line interface"""

    utils.set_logging_format()
    config = conf.Configuration()
    _load_plugins(config)

    parser = build_parser()
    args = parser.parse_args()

    plg.BaseCommand.register_config(config)
    config.init_from_external(args.config_file)
    config.set("logging", args.logging)
    work_manager = plg.WorkManager()
    plg.BaseCommand.args_parsed(args, parser, None, config)
    tlsmate = tm.TlsMate(config=config)
    work_manager.run(tlsmate)


# And now load the plugins which are shipped by default with tlsmate...
from tlsmate.plugins import scan, version  # NOQA



# -*- coding: utf-8 -*-
"""Module scanning for the Heartbleet vulnerability

Refer to CVE-2014-0160, bug in openssl versions, no proper length check
for the heartbeat request. Refer to https://heartbleed.com/
"""
# import basic stuff

# import own stuff
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.tls as tls

# import other stuff


class ScanHeartbleed(plg.Worker):
    name = "heartbleed"
    descr = "scan for Heartbleed vulnerability"
    prio = 41

    def run(self):
        if not hasattr(self.server_profile, "features"):
            hb = tls.HeartbeatState.UNDETERMINED

        else:
            hb = getattr(
                self.server_profile.features,
                "heartbeat",
                tls.HeartbeatState.UNDETERMINED,
            )

        state = tls.HeartbleedStatus.UNDETERMINED
        if hb in (tls.HeartbeatState.FALSE, tls.HeartbeatState.NA):
            state = tls.HeartbleedStatus.NOT_APPLICABLE

        elif hb is tls.HeartbeatState.TRUE:
            values = self.server_profile.get_profile_values(
                tls.Version.all(), full_hs=True
            )
            self.client.init_profile(profile_values=values)
            self.client.profile.heartbeat_mode = tls.HeartbeatMode.PEER_ALLOWED_TO_SEND
            with self.client.create_connection() as conn:
                conn.handshake()
                request = msg.HeartbeatRequest()
                request.payload = b"abc"
                request.payload_length = 4
                request.padding = b""
                conn.send(request)
                try:
                    response = conn.wait(msg.HeartbeatResponse, timeout=2000)
                    if response is not None:
                        state = (
                            tls.HeartbleedStatus.VULNERABLE
                            if response.payload_length == 4
                            else tls.HeartbleedStatus.NOT_VULNERABLE
                        )

                except tls.TlsMsgTimeoutError:
                    state = tls.HeartbleedStatus.TIMEOUT

                except tls.TlsConnectionClosedError:
                    state = tls.HeartbleedStatus.CONNECTION_CLOSED

        self.server_profile.allocate_vulnerabilities()
        self.server_profile.vulnerabilities.heartbleed = state



# -*- coding: utf-8 -*-

# import basic stuff

# import own stuff
from tlsmate import tls

# import other stuff


def test_order_supported_versions(tlsmate):
    tlsmate.client.set_profile(tls.Profile.MODERN)
    tlsmate.client.profile.versions = [
        tls.Version.TLS10,
        tls.Version.SSL20,
        tls.Version.TLS13,
        tls.Version.TLS11,
    ]
    client_hello = tlsmate.client.client_hello()
    versions = client_hello.get_extension(tls.Extension.SUPPORTED_VERSIONS).versions
    assert versions == [
        tls.Version.TLS13,
        tls.Version.TLS11,
        tls.Version.TLS10,
        tls.Version.SSL20,
    ]



# -*- coding: utf-8 -*-
"""Module for CRL manager
"""

# import basic stuff
import logging
from typing import Optional, Dict, List
import datetime

# import own stuff
import tlsmate.cert as crt
import tlsmate.recorder as rec
import tlsmate.cert_utils as cert_utils
import tlsmate.tls as tls

# import other stuff
import requests
from cryptography import x509


class CrlManager(object):
    """Handles all CRL related operations and acts as a cache as well
    """

    def __init__(self, recorder: rec.Recorder) -> None:
        self._crls: Dict[str, Optional[x509.CertificateRevocationList]] = {}
        self._recorder = recorder

    def add_crl(
        self, url: str, der_crl: Optional[bytes] = None, pem_crl: Optional[bytes] = None
    ) -> None:
        """Adds a URL and the CRL to the cache.

        Either der_crl or pem_crl must be given.

        Arguments:
            url: the URL of the CRL
            der_crl: the CRL in DER format given as bytes
            pem_crl: the CRL in PEM format given as bytes
        """

        crl = None
        if der_crl is not None:
            crl = x509.load_der_x509_crl(der_crl)

        elif pem_crl is not None:
            crl = x509.load_pem_x509_crl(pem_crl)

        self._crls[url] = crl

    def _get_crl_obj(self, url, proxies):
        """Get the plain CRL object for a given URL.
        """

        if url not in self._crls:
            bin_crl = None
            self._recorder.trace(crl_url=url)
            try:
                if self._recorder.is_injecting():
                    bin_crl = self._recorder.inject(crl=None)

                else:
                    crl_resp = requests.get(url, timeout=5, proxies=proxies)
                    if crl_resp.ok:
                        bin_crl = crl_resp.content

                    self._recorder.trace(crl=bin_crl)

            except Exception:
                self._crls[url] = None

            else:
                self.add_crl(url, der_crl=bin_crl)

        return self._crls[url]

    def get_crl_status(
        self,
        urls: List[str],
        serial_nbr: int,
        issuer: x509.Name,
        issuer_cert: crt.Certificate,
        timestamp: datetime.datetime,
        proxies: Dict[str, str],
    ) -> Optional[tls.CertCrlStatus]:
        """Determines the CRL revocation status for a given cert/urls.

        Downloads the CRL (if a download fails, the next url is tried), if not yet
        cached, validates the CRL against the issuer & its signature and checks if
        the certificate is present in the CRL or not.

        Arguments:
            urls: a list of CRL-urls
            serial_nbr: the serial number of the certificate to check
            issuer: the issuer name of the cert to check
            issuer_cert: the certificate of the issuer
            timestamp: the time stamp to check the validity of the crl

        Returns:
            the final status.
        """
        status = None
        for url in urls:
            logging.debug(f"downloading CRL from {url}")
            crl = self._get_crl_obj(url, proxies)
            if crl is None:
                status = tls.CertCrlStatus.CRL_DOWNLOAD_FAILED
                continue

            if not cert_utils.equal_names(issuer, issuer_cert.parsed.subject):
                return tls.CertCrlStatus.WRONG_CRL_ISSUER

            if crl.last_update > timestamp or crl.next_update < timestamp:
                return tls.CertCrlStatus.INVALID_TIMESTAMP

            if not crl.is_signature_valid(issuer_cert.parsed.public_key()):
                return tls.CertCrlStatus.CRL_SIGNATURE_INVALID

            if crl.get_revoked_certificate_by_serial_number(serial_nbr) is None:
                return tls.CertCrlStatus.NOT_REVOKED

            else:
                return tls.CertCrlStatus.REVOKED

        return status



# -*- coding: utf-8 -*-
"""Module containing the test suite
"""
# import basic stuff
import abc

# import own stuff
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.server_profile as server_profile
import tlsmate.tls as tls
import tlsmate.utils as utils

# import other stuff


class _Backend(metaclass=abc.ABCMeta):
    @staticmethod
    def get_sig_alg_from_server(client, sig_algs, cert_algs):
        raise NotImplementedError


class _BackendTls12(_Backend):
    @staticmethod
    def get_sig_alg_from_server(client, sig_algs, cert_algs):
        sig_alg = None
        cert_chain = None
        client.profile.signature_algorithms = sig_algs + cert_algs
        with client.create_connection() as conn:
            conn.send(msg.ClientHello)
            conn.wait(msg.ServerHello)
            cert_chain = conn.wait(msg.Certificate).chain
            msg_ske = conn.wait(msg.ServerKeyExchange)
            if msg_ske.ec is not None:
                sig_alg = msg_ske.ec.sig_scheme

            elif msg_ske.dh is not None:
                sig_alg = msg_ske.dh.sig_scheme

        return sig_alg, cert_chain


class _BackendTls13(_Backend):
    @staticmethod
    def get_sig_alg_from_server(client, sig_algs, cert_algs):
        sig_alg = None
        cert_chain = None
        client.profile.signature_algorithms = sig_algs + cert_algs
        with client.create_connection() as conn:
            conn.send(msg.ClientHello)
            conn.wait(msg.ServerHello)
            conn.wait(msg.ChangeCipherSpec, optional=True)
            conn.wait(msg.EncryptedExtensions)
            cert_chain = conn.wait(msg.Certificate).chain
            sig_alg = conn.wait(msg.CertificateVerify).signature_scheme

        return sig_alg, cert_chain


class ScanSigAlgs(plg.Worker):
    name = "sigalgo"
    descr = "scan for supported signature algorithms"
    prio = 20

    def _scan_auth_method(
        self, cipher_suites, sig_algs, cert_algs, prof_sig_algo, backend
    ):
        cert_algs = [alg for alg in cert_algs if alg not in sig_algs]
        sig_alg_supported = []
        if not cipher_suites:
            return sig_alg_supported
        self.client.profile.cipher_suites = cipher_suites
        self.client.profile.signature_algorithms = sig_algs

        while sig_algs:
            sig_alg, cert_chain = backend.get_sig_alg_from_server(
                self.client, sig_algs, cert_algs
            )
            if sig_alg is None or sig_alg in cert_algs:
                break

            self.server_profile.append_unique_cert_chain(cert_chain)
            if sig_alg not in sig_algs:
                if not hasattr(prof_sig_algo, "info"):
                    prof_sig_algo.info = []

                info = f"server selects sig_alg {sig_alg} even when not offered"
                if info not in prof_sig_algo.info:
                    prof_sig_algo.info.append(info)

                break

            sig_alg_supported.append(sig_alg)
            sig_algs.remove(sig_alg)

        for sig_algo in sig_alg_supported:
            prof_sig_algo.algorithms.append(sig_algo)

    def _scan_tls12(self):
        prof_version = self.server_profile.get_version_profile(tls.Version.TLS12)
        if prof_version is None:
            return

        if not hasattr(prof_version, "signature_algorithms"):
            prof_version.signature_algorithms = server_profile.SPSignatureAlgorithms()

        values = self.server_profile.get_profile_values([tls.Version.TLS12])
        self.client.init_profile(profile_values=values)

        prof_sig_algo = prof_version.signature_algorithms
        sigalg_list = tls.SignatureScheme.all()

        rsa_ciphers = utils.filter_cipher_suites(
            values.cipher_suites, key_auth=[tls.KeyAuthentication.RSA]
        )
        rsa_sigalgs = [
            x
            for x in filter(
                lambda alg: (alg.value & 0xFF) == tls.SignatureAlgorithm.RSA.value,
                sigalg_list,
            )
        ]
        cert_sig_algs = self.server_profile.get_cert_sig_algos(
            key_types=[tls.SignatureAlgorithm.RSA]
        )
        self._scan_auth_method(
            rsa_ciphers, rsa_sigalgs, cert_sig_algs, prof_sig_algo, _BackendTls12
        )

        dsa_ciphers = utils.filter_cipher_suites(
            values.cipher_suites, key_auth=[tls.KeyAuthentication.DSS]
        )
        dsa_sigalgs = [
            x
            for x in filter(
                lambda alg: (alg.value & 0xFF) == tls.SignatureAlgorithm.DSA.value,
                sigalg_list,
            )
        ]
        cert_sig_algs = self.server_profile.get_cert_sig_algos(
            key_types=[tls.SignatureAlgorithm.DSA]
        )
        self._scan_auth_method(
            dsa_ciphers, dsa_sigalgs, cert_sig_algs, prof_sig_algo, _BackendTls12
        )

        ecdsa_ciphers = utils.filter_cipher_suites(
            values.cipher_suites, key_algo=[tls.KeyExchangeAlgorithm.ECDHE_ECDSA]
        )
        ecdsa_sigalgs = [
            x
            for x in filter(
                lambda alg: (alg.value & 0xFF) == tls.SignatureAlgorithm.ECDSA.value,
                sigalg_list,
            )
        ]
        ecdsa_sigalgs.extend([tls.SignatureScheme.ED25519, tls.SignatureScheme.ED448])
        cert_sig_algs = self.server_profile.get_cert_sig_algos(
            key_types=[
                tls.SignatureAlgorithm.ECDSA,
                tls.SignatureAlgorithm.ED25519,
                tls.SignatureAlgorithm.ED448,
            ]
        )
        self._scan_auth_method(
            ecdsa_ciphers, ecdsa_sigalgs, cert_sig_algs, prof_sig_algo, _BackendTls12
        )

    def _scan_tls13(self):
        prof_version = self.server_profile.get_version_profile(tls.Version.TLS13)
        if prof_version is None:
            return

        if not hasattr(prof_version, "signature_algorithms"):
            prof_version.signature_algorithms = server_profile.SPSignatureAlgorithms()

        values = self.server_profile.get_profile_values([tls.Version.TLS12])
        self.client.init_profile(profile_values=values)
        prof_sig_algo = prof_version.signature_algorithms
        self.client.profile.supported_groups = prof_version.supported_groups.groups
        self.client.profile.key_shares = tls.SupportedGroups.all_tls13()
        self.client.profile.versions = [tls.Version.TLS13]

        self._scan_auth_method(
            prof_version.ciphers.cipher_suites,
            tls.SignatureScheme.all(),
            self.server_profile.get_cert_sig_algos(),
            prof_sig_algo,
            _BackendTls13,
        )

    def run(self):
        self._scan_tls12()
        self._scan_tls13()
